#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <sys/time.h>

#define SIZE 5242880		/* should be multiple of 128 */
#define ITERATIONS 100

/*
    icc 11: 			3.26GB/s
    gcc 4.2.4
	-O0:			1.34 GB/s
	-O3: 			2.27 GB/s
	-march=nocona		2.27 GB/s
    gcc 4.4.0
	-O3:			2.29 GB/s
	-march=nocona		2.30 GB/s
    asm/SSE			3.32GB/s

*/

#define MEASURE_DECLARE \
  struct timeval tv1; \
  struct timeval tv2; \
  long tv_us;

#define MEASURE_START \
  gettimeofday(&tv1, NULL);

#define MEASURE_STOP \
  gettimeofday(&tv2, NULL);

#define MEASURE_PRINT \
  tv_us = (tv2.tv_sec - tv1.tv_sec)*1000000 + (tv2.tv_usec - tv1.tv_usec); \
  printf("%12s Throughput: %.2lfGB/s (in doubles)\n", __func__, 1000000.*SIZE*sizeof(double)*ITERATIONS/tv_us/1024/1024/1024);

void do_gold(float *flt, double *dbl) {
    int i;
    volatile int j;		// to prevent reduction of the loop

    MEASURE_DECLARE

    MEASURE_START
    for (j = 0; j < ITERATIONS; j++) {
        for (i=0;i<SIZE;i++) flt[i] = dbl[i];
    }
    MEASURE_STOP
    MEASURE_PRINT
}


void do1(float *flt, double *dbl) {
    volatile int j;		// to prevent reduction of the loop
    double *end;

    MEASURE_DECLARE

    MEASURE_START
    end = dbl + SIZE;
    for (j = 0; j < ITERATIONS; j++) {
        __asm__ __volatile__ (
            "push 	%0				\n\t"
            "push 	%1				\n\t"
            "1:						\n\t"
            "movapd 	(%1), %%xmm2			\n\t"
            "add	$16, %1				\n\t"
            "cvtpd2ps 	%%xmm2, %%xmm1			\n\t"
            "movapd 	(%1), %%xmm2			\n\t"
            "add	$16, %1				\n\t"
            "cvtpd2ps 	%%xmm2, %%xmm3			\n\t"
            "shufps	$68, %%xmm3, %%xmm1		\n\t"
            "movntps	%%xmm1, (%0)			\n\t"	// movaps pollutes cache
            "add	$16, %0				\n\t"
            "cmp	%1, %2				\n\t"
            "jg 	1b				\n\t"
            "pop 	%1				\n\t"
            "pop 	%0				\n\t"
    :
    : "p" (flt), "p" (dbl), "r" (end)
        );
    }
    MEASURE_STOP
    MEASURE_PRINT
}


void do2(float *flt, double *dbl) {
    volatile int j;		// to prevent reduction of the loop

    MEASURE_DECLARE

    MEASURE_START

    long max = SIZE*4;	// +16 corresponds to processing of 4 elements
    long val = 0;
    for (j = 0; j < ITERATIONS; j++) {
        __asm__ __volatile__ (
            "push 	%3				\n\t"
            "1:						\n\t"
            "movapd 	(%1,%3,2), %%xmm2		\n\t"
            "cvtpd2ps 	%%xmm2, %%xmm1			\n\t"
            "movapd 	16(%1,%3,2), %%xmm2		\n\t"

            "cvtpd2ps 	%%xmm2, %%xmm3			\n\t"
            "shufps	$68, %%xmm3, %%xmm1		\n\t"
            "movntps	%%xmm1, (%0,%3,1)		\n\t"
            "add	$16, %3				\n\t"
            "cmp	%2, %3				\n\t"
            "jl 	1b				\n\t"
            "pop 	%3				\n\t"
    :
    : "r" (flt), "r" (dbl), "r" (max), "r" (val)
        );
    }
    MEASURE_STOP
    MEASURE_PRINT
}

void do3(float *flt, double *dbl) {
    volatile int j;		// to prevent reduction of the loop

    MEASURE_DECLARE

    MEASURE_START

    long max = SIZE*4;	// +16 corresponds to processing of 4 elements
    long val = max - 16;
    for (j = 0; j < ITERATIONS; j++) {
        __asm__ __volatile__ (
            "push 	%3				\n\t"
            "1:						\n\t"
            "movapd 	(%1,%3,2), %%xmm2		\n\t"
            "cvtpd2ps 	%%xmm2, %%xmm1			\n\t"
            "movapd 	16(%1,%3,2), %%xmm2		\n\t"

            "cvtpd2ps 	%%xmm2, %%xmm3			\n\t"
            "shufps	$68, %%xmm3, %%xmm1		\n\t"

            "movntps	%%xmm1, (%0,%3,1)		\n\t"	// no cache pollution
            "sub	$16, %3				\n\t"
            "jnz 	1b				\n\t"
            "pop 	%3				\n\t"
    :
    : "r" (flt), "r" (dbl), "r" (max), "r" (val)
        );
    }
    MEASURE_STOP
    MEASURE_PRINT
}


void do10(float *flt, double *dbl) {
    volatile int j;		// to prevent reduction of the loop

    MEASURE_DECLARE

    MEASURE_START

    long max = SIZE*4;	// +16 corresponds to processing of 4 elements
    long val = 0;
    for (j = 0; j < ITERATIONS; j++) {
        __asm__ __volatile__ (
            "push 	%3				\n\t"
            "1:						\n\t"
            "movapd 	(%1,%3,2), %%xmm0		\n\t"
            "cvtpd2ps 	%%xmm0, %%xmm8			\n\t"
            "movapd 	16(%1,%3,2), %%xmm1		\n\t"
            "cvtpd2ps 	%%xmm1, %%xmm9			\n\t"
            "movapd 	32(%1,%3,2), %%xmm2		\n\t"
            "cvtpd2ps 	%%xmm2, %%xmm10			\n\t"
            "movapd 	48(%1,%3,2), %%xmm3		\n\t"
            "cvtpd2ps 	%%xmm3, %%xmm11			\n\t"

            "prefetchnta 128(%1,%3,2)			\n\t"

            "movapd 	64(%1,%3,2), %%xmm4		\n\t"
            "cvtpd2ps 	%%xmm4, %%xmm12			\n\t"
            "movapd 	80(%1,%3,2), %%xmm5		\n\t"
            "cvtpd2ps 	%%xmm5, %%xmm13			\n\t"
            "movapd 	96(%1,%3,2), %%xmm6		\n\t"
            "cvtpd2ps 	%%xmm6, %%xmm14			\n\t"
            "movapd 	112(%1,%3,2), %%xmm7		\n\t"
            "cvtpd2ps 	%%xmm7, %%xmm15			\n\t"

            "prefetchnta 192(%1,%3,2)			\n\t"

            "shufps	$68, 	%%xmm9, %%xmm8		\n\t"
            "shufps	$68, 	%%xmm11, %%xmm10	\n\t"
            "shufps	$68, 	%%xmm13, %%xmm12	\n\t"
            "shufps	$68, 	%%xmm15, %%xmm14	\n\t"

            "movntps	%%xmm8, (%0,%3,1)		\n\t"
            "movntps	%%xmm10, 16(%0,%3,1)		\n\t"
            "movntps	%%xmm12, 32(%0,%3,1)		\n\t"
            "movntps	%%xmm14, 48(%0,%3,1)		\n\t"

            "add	$64, %3				\n\t"
            "cmp	%2, %3				\n\t"
            "jl 	1b				\n\t"
            "pop 	%3				\n\t"

            "sfence"
    :
    : "p" (flt), "p" (dbl), "r" (max), "r" (val)
        );
    }
    MEASURE_STOP
    MEASURE_PRINT
}

void do11(float *flt, double *dbl) {
    volatile int j;		// to prevent reduction of the loop

    MEASURE_DECLARE

    MEASURE_START

    long max = SIZE*4;	// +16 corresponds to processing of 4 elements
    long val = -max;
    float *flt2 = flt + SIZE;
    double *dbl2 = dbl + SIZE;

    for (j = 0; j < ITERATIONS; j++) {

        __asm__ __volatile__ (
            "push 	%2				\n\t"
            "1:						\n\t"
            "movapd 	(%1,%2,2), %%xmm0		\n\t"
            "cvtpd2ps 	%%xmm0, %%xmm8			\n\t"
            "movapd 	16(%1,%2,2), %%xmm1		\n\t"
            "cvtpd2ps 	%%xmm1, %%xmm9			\n\t"
            "movapd 	32(%1,%2,2), %%xmm2		\n\t"
            "cvtpd2ps 	%%xmm2, %%xmm10			\n\t"
            "movapd 	48(%1,%2,2), %%xmm3		\n\t"
            "cvtpd2ps 	%%xmm3, %%xmm11			\n\t"

            "movapd 	64(%1,%2,2), %%xmm4		\n\t"
            "cvtpd2ps 	%%xmm4, %%xmm12			\n\t"
            "movapd 	80(%1,%2,2), %%xmm5		\n\t"
            "cvtpd2ps 	%%xmm5, %%xmm13			\n\t"
            "movapd 	96(%1,%2,2), %%xmm6		\n\t"
            "cvtpd2ps 	%%xmm6, %%xmm14			\n\t"
            "movapd 	112(%1,%2,2), %%xmm7		\n\t"
            "cvtpd2ps 	%%xmm7, %%xmm15			\n\t"

            "shufps	$68, 	%%xmm9, %%xmm8		\n\t"
            "movntps	%%xmm8, (%0,%2,1)		\n\t"
            "shufps	$68, 	%%xmm11, %%xmm10	\n\t"
            "movntps	%%xmm10, 16(%0,%2,1)		\n\t"
            "shufps	$68, 	%%xmm13, %%xmm12	\n\t"
            "movntps	%%xmm12, 32(%0,%2,1)		\n\t"
            "shufps	$68, 	%%xmm15, %%xmm14	\n\t"
            "movntps	%%xmm14, 48(%0,%2,1)		\n\t"

            "add	$64, %2				\n\t"
            "jnz 	1b				\n\t"
            "pop 	%2				\n\t"

            "sfence"
    :
    : "p" (flt2), "p" (dbl2), "r" (val)
        );

    }
    MEASURE_STOP
    MEASURE_PRINT
}

void do12(float *flt, double *dbl) {
    volatile int j;		// to prevent reduction of the loop

    MEASURE_DECLARE

    MEASURE_START

    long max = SIZE*4;	// +16 corresponds to processing of 4 elements
    long val = -max;
    float *flt2 = flt + SIZE;
    double *dbl2 = dbl + SIZE;

    for (j = 0; j < ITERATIONS; j++) {

        __asm__ __volatile__ (
            "push 	%2				\n\t"
            "1:						\n\t"
            "movapd 	(%1,%2,2), %%xmm0		\n\t"
            "cvtpd2ps 	%%xmm0, %%xmm8			\n\t"
            "movapd 	16(%1,%2,2), %%xmm1		\n\t"
            "cvtpd2ps 	%%xmm1, %%xmm9			\n\t"
            "movapd 	32(%1,%2,2), %%xmm2		\n\t"
            "cvtpd2ps 	%%xmm2, %%xmm10			\n\t"
            "movapd 	48(%1,%2,2), %%xmm3		\n\t"
            "cvtpd2ps 	%%xmm3, %%xmm11			\n\t"

            "prefetchnta 128(%1,%2,2)			\n\t"

            "movapd 	64(%1,%2,2), %%xmm4		\n\t"
            "cvtpd2ps 	%%xmm4, %%xmm12			\n\t"
            "movapd 	80(%1,%2,2), %%xmm5		\n\t"
            "cvtpd2ps 	%%xmm5, %%xmm13			\n\t"
            "movapd 	96(%1,%2,2), %%xmm6		\n\t"
            "cvtpd2ps 	%%xmm6, %%xmm14			\n\t"
            "movapd 	112(%1,%2,2), %%xmm7		\n\t"
            "cvtpd2ps 	%%xmm7, %%xmm15			\n\t"


            "prefetchnta 192(%1,%2,2)			\n\t"

            "shufps	$68, 	%%xmm9, %%xmm8		\n\t"
            "movntps	%%xmm8, (%0,%2,1)		\n\t"
            "shufps	$68, 	%%xmm11, %%xmm10	\n\t"
            "movntps	%%xmm10, 16(%0,%2,1)		\n\t"
            "shufps	$68, 	%%xmm13, %%xmm12	\n\t"
            "movntps	%%xmm12, 32(%0,%2,1)		\n\t"
            "shufps	$68, 	%%xmm15, %%xmm14	\n\t"
            "movntps	%%xmm14, 48(%0,%2,1)		\n\t"

            "add	$64, %2				\n\t"
            "jnz 	1b				\n\t"
            "pop 	%2				\n\t"

            "sfence"
    :
    : "p" (flt2), "p" (dbl2), "r" (val)
        );

    }
    MEASURE_STOP
    MEASURE_PRINT
}

void do13(float *flt, double *dbl) {
    volatile int j;		// to prevent reduction of the loop

    MEASURE_DECLARE

    MEASURE_START

    long max = SIZE*4;	// +16 corresponds to processing of 4 elements
    long val = -max;
    float *flt2 = flt + SIZE;
    double *dbl2 = dbl + SIZE;

    for (j = 0; j < ITERATIONS; j++) {

        __asm__ __volatile__ (
            "push 	%2				\n\t"
            "1:						\n\t"
            "cvtpd2ps 	(%1,%2,2), %%xmm8		\n\t"
            "cvtpd2ps 	16(%1,%2,2), %%xmm9		\n\t"
            "cvtpd2ps 	32(%1,%2,2), %%xmm10		\n\t"
            "cvtpd2ps 	48(%1,%2,2), %%xmm11		\n\t"

            "prefetchnta 128(%1,%2,2)			\n\t"

            "cvtpd2ps 	64(%1,%2,2), %%xmm12		\n\t"
            "cvtpd2ps 	80(%1,%2,2), %%xmm13		\n\t"
            "cvtpd2ps 	96(%1,%2,2), %%xmm14		\n\t"
            "cvtpd2ps 	112(%1,%2,2), %%xmm15		\n\t"

            "prefetchnta 192(%1,%2,2)			\n\t"

            "shufps	$68, 	%%xmm9, %%xmm8		\n\t"
            "movntps	%%xmm8, (%0,%2,1)		\n\t"
            "shufps	$68, 	%%xmm11, %%xmm10	\n\t"
            "movntps	%%xmm10, 16(%0,%2,1)		\n\t"
            "shufps	$68, 	%%xmm13, %%xmm12	\n\t"
            "movntps	%%xmm12, 32(%0,%2,1)		\n\t"
            "shufps	$68, 	%%xmm15, %%xmm14	\n\t"
            "movntps	%%xmm14, 48(%0,%2,1)		\n\t"

            "add	$64, %2				\n\t"
            "jnz 	1b				\n\t"
            "pop 	%2				\n\t"

            "sfence"
    :
    : "p" (flt2), "p" (dbl2), "r" (val)
        );

    }
    MEASURE_STOP
    MEASURE_PRINT
}

void do14(float *flt, double *dbl) {
    volatile int j;		// to prevent reduction of the loop

    MEASURE_DECLARE

    MEASURE_START

    long max = SIZE*4;	// +16 corresponds to processing of 4 elements
    long val = -max;
    float *flt2 = flt + SIZE;
    double *dbl2 = dbl + SIZE;

    for (j = 0; j < ITERATIONS; j++) {

        __asm__ __volatile__ (
            "push 	%2				\n\t"
            "1:						\n\t"
            "cvtpd2ps 	(%1,%2,2), %%xmm0		\n\t"
            "cvtpd2ps 	16(%1,%2,2), %%xmm1		\n\t"
            "cvtpd2ps 	32(%1,%2,2), %%xmm2		\n\t"
            "cvtpd2ps 	48(%1,%2,2), %%xmm3		\n\t"

            "prefetchnta 128(%1,%2,2)			\n\t"

            "cvtpd2ps 	64(%1,%2,2), %%xmm4		\n\t"
            "cvtpd2ps 	80(%1,%2,2), %%xmm5		\n\t"
            "cvtpd2ps 	96(%1,%2,2), %%xmm6		\n\t"
            "cvtpd2ps 	112(%1,%2,2), %%xmm7		\n\t"

            "prefetchnta 192(%1,%2,2)			\n\t"

            "shufps	$68, 	%%xmm1, %%xmm0		\n\t"
            "shufps	$68, 	%%xmm3, %%xmm2		\n\t"
            "shufps	$68, 	%%xmm5, %%xmm4		\n\t"
            "shufps	$68, 	%%xmm7, %%xmm6		\n\t"

            "movntdq	%%xmm0, (%0,%2,1)		\n\t"
            "movntdq	%%xmm2, 16(%0,%2,1)		\n\t"
            "movntdq	%%xmm4, 32(%0,%2,1)		\n\t"
            "movntdq	%%xmm6, 48(%0,%2,1)		\n\t"
//            "movntdqa	%%xmm0, (%0,%2,1)		\n\t"

            "add	$64, %2				\n\t"
            "jnz 	1b				\n\t"
            "pop 	%2				\n\t"

            "sfence"
    :
    : "p" (flt2), "p" (dbl2), "r" (val)
        );

    }
    MEASURE_STOP
    MEASURE_PRINT
/*
    long mxcsr = 0;
    long *ptr = &mxcsr;
    __asm__ __volatile__ (
	".intel_syntax noprefix		\n\t"
	"stmxcsr [%0]			\n\t"
	 ".att_syntax prefix		\n\t"
	:
	: "r" (ptr)
    );
    printf("%p %p\n", &mxcsr, ptr);
    printf("%lx\n", mxcsr);
*/

}



/*
void do16(float *flt, double *dbl) {
    volatile int j;		// to prevent reduction of the loop

    MEASURE_DECLARE

    MEASURE_START

    long max = SIZE*4;	// +16 corresponds to processing of 4 elements
    long val = -max;
    float *flt2 = flt + SIZE;
    double *dbl2 = dbl + SIZE;

    for (j = 0; j < ITERATIONS; j++) {

        __asm__ __volatile__ (
            "push       %2                             \n\t"
            "1:                                        \n\t"
            "cvtpd2ps   (%1,%2,2), %%xmm0              \n\t"
            "cvtpd2ps   16(%1,%2,2), %%xmm1            \n\t"
            "cvtpd2ps   32(%1,%2,2), %%xmm2            \n\t"
            "cvtpd2ps   48(%1,%2,2), %%xmm3            \n\t"

            "prefetchnta 128(%1,%2,2)                  \n\t"

            "cvtpd2ps   64(%1,%2,2), %%xmm4            \n\t"
            "shufps     $68,    %%xmm1, %%xmm0         \n\t"
            "movntps    %%xmm0, (%0,%2,1)              \n\t"

            "cvtpd2ps   80(%1,%2,2), %%xmm5            \n\t"
            "cvtpd2ps   96(%1,%2,2), %%xmm6            \n\t"
            "shufps     $68,    %%xmm3, %%xmm2         \n\t"
            "movntps    %%xmm2, 16(%0,%2,1)            \n\t"
            "cvtpd2ps   112(%1,%2,2), %%xmm7           \n\t"

            "prefetchnta 192(%1,%2,2)                  \n\t"

            "cvtpd2ps   128(%1,%2,2), %%xmm8           \n\t"
            "shufps     $68,    %%xmm5, %%xmm4         \n\t"
            "movntps    %%xmm4, 32(%0,%2,1)            \n\t"
            "cvtpd2ps   144(%1,%2,2), %%xmm9           \n\t"
            "cvtpd2ps   160(%1,%2,2), %%xmm10          \n\t"
            "shufps     $68,    %%xmm7, %%xmm6         \n\t"
            "movntps    %%xmm6, 48(%0,%2,1)            \n\t"
            "cvtpd2ps   176(%1,%2,2), %%xmm11          \n\t"

            "prefetchnta 256(%1,%2,2)                  \n\t"

            "cvtpd2ps   192(%1,%2,2), %%xmm12          \n\t"
            "shufps     $68,    %%xmm9, %%xmm8         \n\t"
            "movntps    %%xmm8, 64(%0,%2,1)            \n\t"
            "cvtpd2ps   208(%1,%2,2), %%xmm13          \n\t"
            "cvtpd2ps   224(%1,%2,2), %%xmm14          \n\t"
            "shufps     $68,    %%xmm11, %%xmm10       \n\t"
            "movntps    %%xmm10, 80(%0,%2,1)           \n\t"
            "cvtpd2ps   240(%1,%2,2), %%xmm15          \n\t"

            "prefetchnta 320(%1,%2,2)                  \n\t"

            "shufps     $68,    %%xmm13, %%xmm12       \n\t"
            "movntps    %%xmm12, 96(%0,%2,1)           \n\t"
            "shufps     $68,    %%xmm15, %%xmm14       \n\t"
            "movntps    %%xmm14, 112(%0,%2,1)          \n\t"

            "add        $128, %2                       \n\t"
            "jnz        1b                             \n\t"
            "pop        %2                             \n\t"

            "sfence"
    :
    : "p" (flt2), "p" (dbl2), "r" (val)
        );

    }
    MEASURE_STOP
    MEASURE_PRINT
}
*/

/*
void do20(float *flt, double *dbl) {
    volatile int j;		// to prevent reduction of the loop

    MEASURE_DECLARE

    MEASURE_START

    long max = SIZE*4;	// +16 corresponds to processing of 4 elements
    long val = 0;

    for (j = 0; j < ITERATIONS; j++) {

        __asm__ __volatile__ (
            "push 	%2				\n\t"
            "push 	%%rax				\n\t"
            "push 	%%rdx				\n\t"
            "push 	%%rsi				\n\t"
            "1:						\n\t"
            "mov %1, %%rsi				\n\t"

            "add $4096, %%rsi				\n\t"


            "mov $32, %%rax				\n\t"
            "2:						\n\t"
            "mov -64(%%rsi,%2,2), %%rdx			\n\t"
            "mov -128(%%rsi,%2,2), %%rdx		\n\t"
            "sub $128, %%rsi				\n\t"
	    "dec %%rax					\n\t"
	    "jnz 2b					\n\t"


//            "prefetchnta (%1,%2,2)			\n\t"
//            "prefetchnta 64(%1,%2,2)			\n\t"

            "mov $32, %%rax				\n\t"
	    "3:						\n\t"
            "movapd 	(%1,%2,2), %%xmm0		\n\t"
            "cvtpd2ps 	%%xmm0, %%xmm8			\n\t"
            "movapd 	16(%1,%2,2), %%xmm1		\n\t"
            "cvtpd2ps 	%%xmm1, %%xmm9			\n\t"
            "movapd 	32(%1,%2,2), %%xmm2		\n\t"
            "cvtpd2ps 	%%xmm2, %%xmm10			\n\t"
            "movapd 	48(%1,%2,2), %%xmm3		\n\t"
            "cvtpd2ps 	%%xmm3, %%xmm11			\n\t"


            "movapd 	64(%1,%2,2), %%xmm4		\n\t"
            "cvtpd2ps 	%%xmm4, %%xmm12			\n\t"
            "movapd 	80(%1,%2,2), %%xmm5		\n\t"
            "cvtpd2ps 	%%xmm5, %%xmm13			\n\t"
            "movapd 	96(%1,%2,2), %%xmm6		\n\t"
            "cvtpd2ps 	%%xmm6, %%xmm14			\n\t"
            "movapd 	112(%1,%2,2), %%xmm7		\n\t"
            "cvtpd2ps 	%%xmm7, %%xmm15			\n\t"

            "shufps	$68, 	%%xmm9, %%xmm8		\n\t"
            "movntps	%%xmm8, (%0,%2,1)		\n\t"
            "shufps	$68, 	%%xmm11, %%xmm10	\n\t"
            "movntps	%%xmm10, 16(%0,%2,1)		\n\t"
            "shufps	$68, 	%%xmm13, %%xmm12	\n\t"
            "movntps	%%xmm12, 32(%0,%2,1)		\n\t"
            "shufps	$68, 	%%xmm15, %%xmm14	\n\t"
            "movntps	%%xmm14, 48(%0,%2,1)		\n\t"

            "add	$64, %2				\n\t"
	    "dec 	%%rax				\n\t"
	    "jnz 	3b				\n\t"

            "cmp	%3, %2				\n\t"
            "jl 	1b				\n\t"


            "pop 	%%rsi				\n\t"
            "pop 	%%rdx				\n\t"
            "pop 	%%rax				\n\t"
            "pop 	%2				\n\t"

            "sfence"
    :
    : "p" (flt), "p" (dbl), "r" (val), "r" (max)
    : "%rax", "%rdx", "%rsi"
        );

    }
    MEASURE_STOP
    MEASURE_PRINT
}
*/


#define RESET \
 for (i=0;i<SIZE;i++) flt[i] = i + 1;

#define VALIDATE(test) \
 for (i=0;i<SIZE;i++) \
    if (flt[i] != gold[i]) { \
	printf("TEST %i: Error in position: %i, %lf (test) != %lf (gold)\n", test, i, flt[i], gold[i]); \
	exit(1); \
    }

#define TEST(i) \
 RESET \
 do##i(flt, dbl); \
 VALIDATE(i) \
 
int main() {
    int i;
    float *flt, *gold;
    double *dbl;

    gold = (float*)malloc((SIZE+128)*sizeof(float));
    flt = (float*)malloc((SIZE+128)*sizeof(float));
    dbl = (double*)malloc((SIZE+128)*sizeof(double));

    for (i=0;i<SIZE;i++) dbl[i] = i + 1; // 1e-200;

    do_gold(gold, dbl);


//    TEST(1)
    TEST(2)
//    TEST(3)


    TEST(10)
    TEST(11)
    TEST(12)
    TEST(13)

    TEST(14)

    free(dbl);
    free(flt);
    free(gold);
}
