#define _XOPEN_SOURCE 600
#include <stdlib.h>

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <inttypes.h>

#include <pthread.h>
#include <omp.h>
#include <sys/time.h>

#if MSVCPP
# define ALGNW __declspec(align(16))
# define ALGNL
#else
# define ALGNW
# define ALGNL __attribute__((aligned(16)))
#endif


#define BLOCK 1024
#define SIZE 5242880		/* should be multiple of 128 */
#define ITERATIONS 100

/*
    icc 11: 			0.18 GB/s
    gcc 4.2.4
	-O3: 			0.29 GB/s
    Optimized:			1.14 GB/s
    SSE/DO2			2.90 GB/s
    SSE/PMULHRSW		3.35 GB/s
    SSE/Unloop2+prefetch	3.50 GB/s
    SSE/Unloop4			slightly slower
    SSE/Max			3.51 GB/s
    SSE/Threaded		3.65 GB/s
*/


int num_procs;

#define max2(a,b) (((a)>(b))?(a):(b))
#define min2(a,b) (((a)<(b))?(a):(b))

#define calc_alloc(size,rounding) ((((size)/(rounding)) + (((size)%(rounding))?1:0))*(rounding))
#define calc_blocks(size,rounding) (((size)/(rounding)) + (((size)%(rounding))?1:0))

#define MEASURE_DECLARE \
  struct timeval tv1; \
  struct timeval tv2; \
  long tv_us;

#define MEASURE_START \
  gettimeofday(&tv1, NULL);

#define MEASURE_STOP \
  gettimeofday(&tv2, NULL);

#define MEASURE_PRINT \
  tv_us = (tv2.tv_sec - tv1.tv_sec)*1000000 + (tv2.tv_usec - tv1.tv_usec); \
  printf("%12s Throughput: %.2lfGB/s (basing on source size)\n", __func__, 1000000.*SIZE*sizeof(uint32_t)*ITERATIONS/tv_us/1024/1024/1024);

#define MEASURE_PRINT_PROCS \
  tv_us = (tv2.tv_sec - tv1.tv_sec)*1000000 + (tv2.tv_usec - tv1.tv_usec); \
  printf("%12s (% 3u threads) Throughput: %.2lfGB/s (basing on source size)\n", __func__, procs, 1000000.*SIZE*sizeof(uint32_t)*ITERATIONS/tv_us/1024/1024/1024);


void do_gold(unsigned char *dst, unsigned char *src) {
    int i;
    volatile int j;		// to prevent reduction of the loop

    MEASURE_DECLARE

    MEASURE_START
    for (j = 0; j < ITERATIONS; j++) {
        for (i=0;i<SIZE;i++) {
	    dst[i] = roundf(((float)(src[4*i] + src[4*i+1] + src[4*i+2]))/3);
	}
    }
    MEASURE_STOP
    MEASURE_PRINT
}
unsigned char *gold;
void do1(unsigned char *dst, unsigned char *src) {
    int i;
    volatile int j;		// to prevent reduction of the loop
    unsigned short sum;

	// We just approximating division by a siries of rational
	// divisions i.e. a / 3 ~ 85*a/256 (fits in 85*a fits in word).
	// Two such steps is enough for numbers below 256. The added
	// 128 is two emulate round ap to closest:
	// a/3 is 0, 0.33, or 0.66, therefore for desired rounding mode
	// we should add to a a number between 0.34 and 0.77.

    MEASURE_DECLARE
    MEASURE_START
    for (j = 0; j < ITERATIONS; j++) {
        for (i=0;i<SIZE;i++) {
	    sum = (src[4*i] + src[4*i+1] + src[4*i+2]);
	    dst[i] = (90 + 85 * sum + ((85 * sum)>>8)) >> 8;
//		No performance difference with the above line
//	    dst[i] = (45 + 85 * sum + (85 * sum)/256) / 256;
//	    dst[i] = (65 + 85 * sum) / 256;
	}
    }
    MEASURE_STOP
    MEASURE_PRINT
}


void do2(unsigned char *dst, unsigned char *src) {
    int i;
    volatile int j;		// to prevent reduction of the loop
    unsigned char *end;

    MEASURE_DECLARE

    MEASURE_START
    end = dst + SIZE;
    
    ALGNW char shuf[80] ALGNL;
    char *shuf1 = shuf;
    char *shuf2 = shuf+16;
    char *shuf3 = shuf+32;
    uint16_t *mul = (uint16_t*)(shuf+48);
    uint16_t *add = (uint16_t*)(shuf+64);
//    printf("%p!!!\n", shuf);

    memset(shuf, 0xFF, 48);
    
    for (i=0;i<16;i+=4) {
	shuf1[i] = i;
//	shuf1[i+1] = 0;
	shuf1[i+2] = i + 1;
//	shuf1[i+3] = 0;
    }

    for (i=0;i<8;i+=2) {
	shuf2[i] = 2 * i + 2;
//	shuf2[i+1] = 0;
	
	shuf3[i + 8] = 2 * i + 2;
//	shuf3[i + 9] = 0;
    }
    for (i=0;i<8;i++) {
	mul[i] = 85;
	add[i] = 90;
    }
    
    for (j = 0; j < ITERATIONS; j++) {
        __asm__ __volatile__ (
            "push 	%0				\n\t"
            "push 	%1				\n\t"
	    "movdqa	(%3), %%xmm15			\n\t"
	    "movdqa	16(%3), %%xmm14			\n\t"
	    "movdqa	32(%3), %%xmm13			\n\t"
	    "movdqa	48(%3), %%xmm12			\n\t"
	    "movdqa	64(%3), %%xmm11			\n\t"
            "1:						\n\t"
//          "movntdqa	(%0), %%xmm6			\n\t"
            "movdqa	(%1), %%xmm0			\n\t"
            "movdqa	16(%1), %%xmm1			\n\t"
	    
	    "movdqa	%%xmm0, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm1, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm0			\n\t"
	    "pshufb	%%xmm15, %%xmm1			\n\t"
	    "phaddw	%%xmm1, %%xmm0			\n\t"
	    "paddw	%%xmm5, %%xmm0			\n\t"

	    "pmullw	%%xmm12,%%xmm0			\n\t"
	    "movdqa	%%xmm0, %%xmm1			\n\t"
	    "psrlw	$0x08, %%xmm0			\n\t"
	    "paddw	%%xmm1, %%xmm0			\n\t"
	    "paddw	%%xmm11, %%xmm0			\n\t"
	    "psrlw	$0x08, %%xmm0			\n\t"
    
//------
            "movdqa	32(%1), %%xmm1			\n\t"
            "movdqa	48(%1), %%xmm2			\n\t"
	    
	    "movdqa	%%xmm1, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm2, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm1			\n\t"
	    "pshufb	%%xmm15, %%xmm2			\n\t"
	    "phaddw	%%xmm2, %%xmm1			\n\t"
	    "paddw	%%xmm5, %%xmm1			\n\t"

	    "pmullw	%%xmm12,%%xmm1			\n\t"
	    "movdqa	%%xmm1, %%xmm2			\n\t"
	    "psrlw	$0x08, %%xmm1			\n\t"
	    "paddw	%%xmm2, %%xmm1			\n\t"
	    "paddw	%%xmm11, %%xmm1			\n\t"
	    "psrlw	$0x08, %%xmm1			\n\t"
//-----
	    "packuswb	%%xmm1, %%xmm0			\n\t"
            "movntps	%%xmm0, (%0)			\n\t"

            "add	$16, %0				\n\t"
            "add	$64, %1				\n\t"
            "cmp	%0, %2				\n\t"
            "jg 	1b				\n\t"
            "pop 	%1				\n\t"
            "pop 	%0				\n\t"

    :
    : "p" (dst), "p" (src), "r" (end), "p" (&shuf)
    );
	}
    MEASURE_STOP
    MEASURE_PRINT
/*
    for (i=0;i<16;i++) {
	if ((i%4)==0) printf("\n");
	printf("%3i  %3i  %3i       ", src[4*i], src[4*i+1], src[4*i+2]);
    }
    printf("\n\n");

    for (i=0;i<16;i++) {
	if ((i%4)==0) printf("\n");
	printf("%3x  ", dst[i]);
    }
    printf("\n");
*/
}

void do3(unsigned char *dst, unsigned char *src) {
    int i;
    volatile int j;		// to prevent reduction of the loop
    unsigned char *end;

    MEASURE_DECLARE

    MEASURE_START
    end = dst + SIZE;
    
    ALGNW char shuf[64] ALGNL;
    char *shuf1 = shuf;
    char *shuf2 = shuf+16;
    char *shuf3 = shuf+32;
    uint16_t *mul = (uint16_t*)(shuf+48);

    memset(shuf, 0xFF, 48);
    
    for (i=0;i<16;i+=4) {
	shuf1[i] = i;
//	shuf1[i+1] = 0;
	shuf1[i+2] = i + 1;
//	shuf1[i+3] = 0;
    }

    for (i=0;i<8;i+=2) {
	shuf2[i] = 2 * i + 2;
//	shuf2[i+1] = 0;
	
	shuf3[i + 8] = 2 * i + 2;
//	shuf3[i + 9] = 0;
    }
    for (i=0;i<8;i++) {
	mul[i] = 0x2AAA;
    }
    
    for (j = 0; j < ITERATIONS; j++) {
        __asm__ __volatile__ (
            "push 	%0				\n\t"
            "push 	%1				\n\t"
	    "movdqa	(%3), %%xmm15			\n\t"
	    "movdqa	16(%3), %%xmm14			\n\t"
	    "movdqa	32(%3), %%xmm13			\n\t"
	    "movdqa	48(%3), %%xmm12			\n\t"
            "1:						\n\t"
//          "movntdqa	(%0), %%xmm6			\n\t"
            "movdqa	(%1), %%xmm0			\n\t"
            "movdqa	16(%1), %%xmm1			\n\t"
	    
	    "movdqa	%%xmm0, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm1, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm0			\n\t"
	    "pshufb	%%xmm15, %%xmm1			\n\t"
	    "phaddw	%%xmm1, %%xmm0			\n\t"
	    "paddw	%%xmm5, %%xmm0			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm0			\n\t"
//------
            "movdqa	32(%1), %%xmm1			\n\t"
            "movdqa	48(%1), %%xmm2			\n\t"
	    
	    "movdqa	%%xmm1, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm2, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm1			\n\t"
	    "pshufb	%%xmm15, %%xmm2			\n\t"
	    "phaddw	%%xmm2, %%xmm1			\n\t"
	    "paddw	%%xmm5, %%xmm1			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm1			\n\t"

//-----
	    "packuswb	%%xmm1, %%xmm0			\n\t"
            "movntps	%%xmm0, (%0)			\n\t"

            "add	$16, %0				\n\t"
            "add	$64, %1				\n\t"
            "cmp	%0, %2				\n\t"
            "jg 	1b				\n\t"
            "pop 	%1				\n\t"
            "pop 	%0				\n\t"

    :
    : "p" (dst), "p" (src), "r" (end), "p" (&shuf)
    );
	}
    MEASURE_STOP
    MEASURE_PRINT
}

void do4(unsigned char *dst, unsigned char *src) {
    int i;
    volatile int j;		// to prevent reduction of the loop
    unsigned char *end;

    MEASURE_DECLARE

    MEASURE_START
    end = dst + SIZE;
    
    ALGNW char shuf[64] ALGNL;
    char *shuf1 = shuf;
    char *shuf2 = shuf+16;
    char *shuf3 = shuf+32;
    uint16_t *mul = (uint16_t*)(shuf+48);

    memset(shuf, 0xFF, 48);
    
    for (i=0;i<16;i+=4) {
	shuf1[i] = i;
//	shuf1[i+1] = 0;
	shuf1[i+2] = i + 1;
//	shuf1[i+3] = 0;
    }

    for (i=0;i<8;i+=2) {
	shuf2[i] = 2 * i + 2;
//	shuf2[i+1] = 0;
	
	shuf3[i + 8] = 2 * i + 2;
//	shuf3[i + 9] = 0;
    }
    for (i=0;i<8;i++) {
	mul[i] = 0x2AAA;
    }
    
    for (j = 0; j < ITERATIONS; j++) {
        __asm__ __volatile__ (
            "prefetchnta (%1)				\n\t"
            "push 	%0				\n\t"
            "push 	%1				\n\t"
	    "movdqa	(%3), %%xmm15			\n\t"
	    "movdqa	16(%3), %%xmm14			\n\t"
	    "movdqa	32(%3), %%xmm13			\n\t"
	    "movdqa	48(%3), %%xmm12			\n\t"
            "1:						\n\t"
            "movdqa	(%1), %%xmm0			\n\t"
            "movdqa	16(%1), %%xmm1			\n\t"
            "prefetchnta 64(%1)				\n\t"
	    
	    "movdqa	%%xmm0, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm1, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm0			\n\t"
	    "pshufb	%%xmm15, %%xmm1			\n\t"
	    "phaddw	%%xmm1, %%xmm0			\n\t"
	    "paddw	%%xmm5, %%xmm0			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm0			\n\t"
//------
            "movdqa	32(%1), %%xmm1			\n\t"
            "movdqa	48(%1), %%xmm2			\n\t"
	    
	    "movdqa	%%xmm1, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm2, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm1			\n\t"
	    "pshufb	%%xmm15, %%xmm2			\n\t"
	    "phaddw	%%xmm2, %%xmm1			\n\t"
	    "paddw	%%xmm5, %%xmm1			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm1			\n\t"

//-----
            "movdqa	64(%1), %%xmm2			\n\t"
            "movdqa	80(%1), %%xmm3			\n\t"
            "prefetchnta 128(%1)			\n\t"
	    
	    "movdqa	%%xmm2, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm3, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm2			\n\t"
	    "pshufb	%%xmm15, %%xmm3			\n\t"
	    "phaddw	%%xmm3, %%xmm2			\n\t"
	    "paddw	%%xmm5, %%xmm2			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm2			\n\t"
//------
            "movdqa	96(%1), %%xmm3			\n\t"
            "movdqa	112(%1), %%xmm4			\n\t"
	    
	    "movdqa	%%xmm3, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm4, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm3			\n\t"
	    "pshufb	%%xmm15, %%xmm4			\n\t"
	    "phaddw	%%xmm4, %%xmm3			\n\t"
	    "paddw	%%xmm5, %%xmm3			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm3			\n\t"
//-----

	    "packuswb	%%xmm1, %%xmm0			\n\t"
	    "packuswb	%%xmm3, %%xmm2			\n\t"
            "movntps	%%xmm0, (%0)			\n\t"
            "movntps	%%xmm2, 16(%0)			\n\t"

            "add	$32, %0				\n\t"
            "add	$128, %1			\n\t"
            "cmp	%0, %2				\n\t"
            "jg 	1b				\n\t"
            "pop 	%1				\n\t"
            "pop 	%0				\n\t"

    :
    : "p" (dst), "p" (src), "r" (end), "p" (&shuf)
    );
	}
    MEASURE_STOP
    MEASURE_PRINT
}

void do5(unsigned char *dst, unsigned char *src) {
    int i;
    volatile int j;		// to prevent reduction of the loop
    unsigned char *end;

    MEASURE_DECLARE

    MEASURE_START
    end = dst + SIZE;
    
    ALGNW char shuf[64] ALGNL;
    char *shuf1 = shuf;
    char *shuf2 = shuf+16;
    char *shuf3 = shuf+32;
    uint16_t *mul = (uint16_t*)(shuf+48);

    memset(shuf, 0xFF, 48);
    
    for (i=0;i<16;i+=4) {
	shuf1[i] = i;
//	shuf1[i+1] = 0;
	shuf1[i+2] = i + 1;
//	shuf1[i+3] = 0;
    }

    for (i=0;i<8;i+=2) {
	shuf2[i] = 2 * i + 2;
//	shuf2[i+1] = 0;
	
	shuf3[i + 8] = 2 * i + 2;
//	shuf3[i + 9] = 0;
    }
    for (i=0;i<8;i++) {
	mul[i] = 0x2AAA;
    }
    
    for (j = 0; j < ITERATIONS; j++) {
        __asm__ __volatile__ (
            "prefetchnta (%1)				\n\t"
            "push 	%0				\n\t"
            "push 	%1				\n\t"
	    "movdqa	(%3), %%xmm15			\n\t"
	    "movdqa	16(%3), %%xmm14			\n\t"
	    "movdqa	32(%3), %%xmm13			\n\t"
	    "movdqa	48(%3), %%xmm12			\n\t"
            "1:						\n\t"
            "movdqa	(%1), %%xmm0			\n\t"
            "movdqa	16(%1), %%xmm1			\n\t"
            "prefetchnta 64(%1)				\n\t"
	    
	    "movdqa	%%xmm0, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm1, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm0			\n\t"
	    "pshufb	%%xmm15, %%xmm1			\n\t"
	    "phaddw	%%xmm1, %%xmm0			\n\t"
	    "paddw	%%xmm5, %%xmm0			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm0			\n\t"
//------
            "movdqa	32(%1), %%xmm1			\n\t"
            "movdqa	48(%1), %%xmm2			\n\t"
	    
	    "movdqa	%%xmm1, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm2, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm1			\n\t"
	    "pshufb	%%xmm15, %%xmm2			\n\t"
	    "phaddw	%%xmm2, %%xmm1			\n\t"
	    "paddw	%%xmm5, %%xmm1			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm1			\n\t"

//-----
            "movdqa	64(%1), %%xmm2			\n\t"
            "movdqa	80(%1), %%xmm3			\n\t"
            "prefetchnta 128(%1)			\n\t"
	    
	    "movdqa	%%xmm2, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm3, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm2			\n\t"
	    "pshufb	%%xmm15, %%xmm3			\n\t"
	    "phaddw	%%xmm3, %%xmm2			\n\t"
	    "paddw	%%xmm5, %%xmm2			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm2			\n\t"
//------
            "movdqa	96(%1), %%xmm3			\n\t"
            "movdqa	112(%1), %%xmm4			\n\t"
	    
	    "movdqa	%%xmm3, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm4, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm3			\n\t"
	    "pshufb	%%xmm15, %%xmm4			\n\t"
	    "phaddw	%%xmm4, %%xmm3			\n\t"
	    "paddw	%%xmm5, %%xmm3			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm3			\n\t"
//-----
//-----


            "movdqa	128(%1), %%xmm7			\n\t"
            "movdqa	144(%1), %%xmm8			\n\t"
            "prefetchnta 192(%1)			\n\t"
	    
	    "movdqa	%%xmm7, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm8, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm7			\n\t"
	    "pshufb	%%xmm15, %%xmm8			\n\t"
	    "phaddw	%%xmm8, %%xmm7			\n\t"
	    "paddw	%%xmm5, %%xmm7			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm7			\n\t"
//------
            "movdqa	160(%1), %%xmm8			\n\t"
            "movdqa	176(%1), %%xmm9			\n\t"
	    
	    "movdqa	%%xmm8, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm9, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm8			\n\t"
	    "pshufb	%%xmm15, %%xmm9			\n\t"
	    "phaddw	%%xmm9, %%xmm8			\n\t"
	    "paddw	%%xmm5, %%xmm8			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm8			\n\t"

//-----
            "movdqa	192(%1), %%xmm9			\n\t"
            "movdqa	208(%1), %%xmm10		\n\t"
            "prefetchnta 256(%1)			\n\t"
	    
	    "movdqa	%%xmm9, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm10, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm9			\n\t"
	    "pshufb	%%xmm15, %%xmm10		\n\t"
	    "phaddw	%%xmm10, %%xmm9			\n\t"
	    "paddw	%%xmm5, %%xmm9			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm9			\n\t"
//------
            "movdqa	224(%1), %%xmm10		\n\t"
            "movdqa	240(%1), %%xmm11		\n\t"
	    
	    "movdqa	%%xmm10, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm11, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm10		\n\t"
	    "pshufb	%%xmm15, %%xmm11		\n\t"
	    "phaddw	%%xmm11, %%xmm10		\n\t"
	    "paddw	%%xmm5, %%xmm10			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm10		\n\t"

//------
	    "packuswb	%%xmm1, %%xmm0			\n\t"
	    "packuswb	%%xmm3, %%xmm2			\n\t"
	    "packuswb	%%xmm8, %%xmm7			\n\t"
	    "packuswb	%%xmm10, %%xmm9			\n\t"
//------

            "movntps	%%xmm0, (%0)			\n\t"
            "movntps	%%xmm2, 16(%0)			\n\t"
            "movntps	%%xmm7, 32(%0)			\n\t"
            "movntps	%%xmm9, 48(%0)			\n\t"

            "add	$64, %0				\n\t"
            "add	$256, %1			\n\t"
            "cmp	%0, %2				\n\t"
            "jg 	1b				\n\t"
            "pop 	%1				\n\t"
            "pop 	%0				\n\t"

    :
    : "p" (dst), "p" (src), "r" (end), "p" (&shuf)
    );
	}
    MEASURE_STOP
    MEASURE_PRINT
}




void do6(unsigned char *dst, unsigned char *src) {
    int i;
    volatile int j;		// to prevent reduction of the loop
    unsigned char *end;

    MEASURE_DECLARE

    MEASURE_START
    end = dst + SIZE;
    
    ALGNW char shuf[64] ALGNL;
    char *shuf1 = shuf;
    char *shuf2 = shuf+16;
    char *shuf3 = shuf+32;
    uint16_t *mul = (uint16_t*)(shuf+48);

    memset(shuf, 0xFF, 48);
    
    for (i=0;i<16;i+=4) {
	shuf1[i] = i;
//	shuf1[i+1] = 0;
	shuf1[i+2] = i + 1;
//	shuf1[i+3] = 0;
    }

    for (i=0;i<8;i+=2) {
	shuf2[i] = 2 * i + 2;
//	shuf2[i+1] = 0;
	
	shuf3[i + 8] = 2 * i + 2;
//	shuf3[i + 9] = 0;
    }
    for (i=0;i<8;i++) {
	mul[i] = 0x2AAA;
    }

    long max = SIZE;
    long val = -max;
    char *src2 = src + 4 * SIZE;
    char *dst2 = dst + SIZE;
    
    for (j = 0; j < ITERATIONS; j++) {
        __asm__ __volatile__ (
            "prefetchnta (%1)				\n\t"
            "push 	%2				\n\t"
	    "movdqa	(%3), %%xmm15			\n\t"
	    "movdqa	16(%3), %%xmm14			\n\t"
	    "movdqa	32(%3), %%xmm13			\n\t"
	    "movdqa	48(%3), %%xmm12			\n\t"
            "1:						\n\t"
            "movdqa	(%1,%2,4), %%xmm0		\n\t"
            "movdqa	16(%1,%2,4), %%xmm1		\n\t"
        
	    "prefetchnta 64(%1)				\n\t"
	    
	    "movdqa	%%xmm0, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm1, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm0			\n\t"
	    "pshufb	%%xmm15, %%xmm1			\n\t"
	    "phaddw	%%xmm1, %%xmm0			\n\t"
	    "paddw	%%xmm5, %%xmm0			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm0			\n\t"
//------
            "movdqa	32(%1,%2,4), %%xmm1		\n\t"
            "movdqa	48(%1,%2,4), %%xmm2		\n\t"
	    
	    "movdqa	%%xmm1, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm2, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm1			\n\t"
	    "pshufb	%%xmm15, %%xmm2			\n\t"
	    "phaddw	%%xmm2, %%xmm1			\n\t"
	    "paddw	%%xmm5, %%xmm1			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm1			\n\t"

//-----
            "movdqa	64(%1,%2,4), %%xmm2		\n\t"
            "movdqa	80(%1,%2,4), %%xmm3		\n\t"

            "prefetchnta 128(%1)			\n\t"
	    
	    "movdqa	%%xmm2, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm3, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm2			\n\t"
	    "pshufb	%%xmm15, %%xmm3			\n\t"
	    "phaddw	%%xmm3, %%xmm2			\n\t"
	    "paddw	%%xmm5, %%xmm2			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm2			\n\t"
//------
            "movdqa	96(%1,%2,4), %%xmm3		\n\t"
            "movdqa	112(%1,%2,4), %%xmm4		\n\t"
	    
	    "movdqa	%%xmm3, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm4, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm3			\n\t"
	    "pshufb	%%xmm15, %%xmm4			\n\t"
	    "phaddw	%%xmm4, %%xmm3			\n\t"
	    "paddw	%%xmm5, %%xmm3			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm3			\n\t"
//-----

	    "packuswb	%%xmm1, %%xmm0			\n\t"
	    "packuswb	%%xmm3, %%xmm2			\n\t"
            "movntps	%%xmm0, (%0,%2)			\n\t"
            "movntps	%%xmm2, 16(%0,%2)		\n\t"

            "add	$32, %2				\n\t"
            "jnz 	1b				\n\t"
            "pop 	%2				\n\t"

    :
    : "p" (dst2), "p" (src2), "r" (val), "p" (&shuf)
    );
	}
    MEASURE_STOP
    MEASURE_PRINT
}

void do7(unsigned char *dst, unsigned char *src) {
    int i;
    volatile int j;		// to prevent reduction of the loop
    unsigned char *end;

    MEASURE_DECLARE

    MEASURE_START
    end = dst + SIZE;
    
    ALGNW char shuf[64] ALGNL;
    char *shuf1 = shuf;
    char *shuf2 = shuf+16;
    char *shuf3 = shuf+32;
    uint16_t *mul = (uint16_t*)(shuf+48);

    memset(shuf, 0xFF, 48);
    
    for (i=0;i<16;i+=4) {
	shuf1[i] = i;
//	shuf1[i+1] = 0;
	shuf1[i+2] = i + 1;
//	shuf1[i+3] = 0;
    }

    for (i=0;i<8;i+=2) {
	shuf2[i] = 2 * i + 2;
//	shuf2[i+1] = 0;
	
	shuf3[i + 8] = 2 * i + 2;
//	shuf3[i + 9] = 0;
    }
    for (i=0;i<8;i++) {
	mul[i] = 0x2AAA;
    }

    long max = SIZE;
    long val = -max;
    char *src2 = src + 4 * SIZE;
    char *dst2 = dst + SIZE;
    
    for (j = 0; j < ITERATIONS; j++) {
        __asm__ __volatile__ (
            "prefetchnta (%1)				\n\t"
            "push 	%2				\n\t"
	    "movdqa	(%3), %%xmm15			\n\t"
	    "movdqa	16(%3), %%xmm14			\n\t"
	    "movdqa	32(%3), %%xmm13			\n\t"
	    "movdqa	48(%3), %%xmm12			\n\t"
            "1:						\n\t"
            "movdqa	(%1,%2,4), %%xmm0		\n\t"
            "movdqa	16(%1,%2,4), %%xmm1		\n\t"
        
	    "prefetchnta 64(%1)				\n\t"
	    
//	    "movdqa	%%xmm0, %%xmm5			\n\t"
	    "pshufd	$0xE4, %%xmm0, %%xmm5		\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
//	    "movdqa	%%xmm1, %%xmm6			\n\t"
	    "pshufd	$0xE4, %%xmm1, %%xmm6		\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm0			\n\t"
	    "pshufb	%%xmm15, %%xmm1			\n\t"
	    "phaddw	%%xmm1, %%xmm0			\n\t"
	    "paddw	%%xmm5, %%xmm0			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm0			\n\t"
//------
            "movdqa	32(%1,%2,4), %%xmm1		\n\t"
            "movdqa	48(%1,%2,4), %%xmm2		\n\t"
	    
//	    "movdqa	%%xmm1, %%xmm5			\n\t"
	    "pshufd	$0xE4, %%xmm1, %%xmm5		\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
//	    "movdqa	%%xmm2, %%xmm6			\n\t"
	    "pshufd	$0xE4, %%xmm2, %%xmm6		\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm1			\n\t"
	    "pshufb	%%xmm15, %%xmm2			\n\t"
	    "phaddw	%%xmm2, %%xmm1			\n\t"
	    "paddw	%%xmm5, %%xmm1			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm1			\n\t"

//-----
            "movdqa	64(%1,%2,4), %%xmm2		\n\t"
            "movdqa	80(%1,%2,4), %%xmm3		\n\t"

            "prefetchnta 128(%1)			\n\t"
	    
//	    "movdqa	%%xmm2, %%xmm5			\n\t"
	    "pshufd	$0xE4, %%xmm2, %%xmm5		\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
//	    "movdqa	%%xmm3, %%xmm6			\n\t"
	    "pshufd	$0xE4, %%xmm3, %%xmm6		\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm2			\n\t"
	    "pshufb	%%xmm15, %%xmm3			\n\t"
	    "phaddw	%%xmm3, %%xmm2			\n\t"
	    "paddw	%%xmm5, %%xmm2			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm2			\n\t"
//------
            "movdqa	96(%1,%2,4), %%xmm3		\n\t"
            "movdqa	112(%1,%2,4), %%xmm4		\n\t"
	    
//	    "movdqa	%%xmm3, %%xmm5			\n\t"
	    "pshufd	$0xE4, %%xmm3, %%xmm5		\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
//	    "movdqa	%%xmm4, %%xmm6			\n\t"
	    "pshufd	$0xE4, %%xmm4, %%xmm6		\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm3			\n\t"
	    "pshufb	%%xmm15, %%xmm4			\n\t"
	    "phaddw	%%xmm4, %%xmm3			\n\t"
	    "paddw	%%xmm5, %%xmm3			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm3			\n\t"
//-----

	    "packuswb	%%xmm1, %%xmm0			\n\t"
	    "packuswb	%%xmm3, %%xmm2			\n\t"
            "movntps	%%xmm0, (%0,%2)			\n\t"
            "movntps	%%xmm2, 16(%0,%2)		\n\t"

            "add	$32, %2				\n\t"
            "jnz 	1b				\n\t"
            "pop 	%2				\n\t"

    :
    : "p" (dst2), "p" (src2), "r" (val), "p" (&shuf)
    );
	}
    MEASURE_STOP
    MEASURE_PRINT
}



void do8(unsigned char *dst, unsigned char *src) {
    int i;
    volatile int j;		// to prevent reduction of the loop
    unsigned char *end;

    MEASURE_DECLARE

    MEASURE_START
    end = dst + SIZE;
    
    ALGNW char shuf[64] ALGNL;
    char *shuf1 = shuf;
    char *shuf2 = shuf+16;
    char *shuf3 = shuf+32;
    uint16_t *mul = (uint16_t*)(shuf+48);

    memset(shuf, 0xFF, 48);
    
    for (i=0;i<16;i+=4) {
	shuf1[i] = i;
//	shuf1[i+1] = 0;
	shuf1[i+2] = i + 1;
//	shuf1[i+3] = 0;
    }

    for (i=0;i<8;i+=2) {
	shuf2[i] = 2 * i + 2;
//	shuf2[i+1] = 0;
	
	shuf3[i + 8] = 2 * i + 2;
//	shuf3[i + 9] = 0;
    }
    for (i=0;i<8;i++) {
	mul[i] = 0x2AAA;
    }

    long max = SIZE;
    long val = -max;
    char *src2 = src + 4 * SIZE;
    char *dst2 = dst + SIZE;
    
    for (j = 0; j < ITERATIONS; j++) {
        __asm__ __volatile__ (
            "prefetchnta (%1)				\n\t"
            "push 	%2				\n\t"
	    "movdqa	(%3), %%xmm15			\n\t"
	    "movdqa	16(%3), %%xmm14			\n\t"
	    "movdqa	32(%3), %%xmm13			\n\t"
	    "movdqa	48(%3), %%xmm12			\n\t"
            "1:						\n\t"
            "movdqa	(%1,%2,4), %%xmm0		\n\t"
            "movdqa	16(%1,%2,4), %%xmm1		\n\t"
        
	    "prefetchnta 64(%1)				\n\t"
	    
	    "movdqa	%%xmm0, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm1, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

            "movdqa	32(%1,%2,4), %%xmm2		\n\t"
            "movdqa	48(%1,%2,4), %%xmm3		\n\t"

	    "pshufb	%%xmm15, %%xmm0			\n\t"
	    "pshufb	%%xmm15, %%xmm1			\n\t"
	    "phaddw	%%xmm1, %%xmm0			\n\t"
	    "paddw	%%xmm5, %%xmm0			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm0			\n\t"
//------
	    
	    "movdqa	%%xmm2, %%xmm5			\n\t"
	    "movdqa	%%xmm3, %%xmm6			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

            "movdqa	64(%1,%2,4), %%xmm7		\n\t"
            "movdqa	80(%1,%2,4), %%xmm8		\n\t"

	    "pshufb	%%xmm15, %%xmm2			\n\t"
	    "pshufb	%%xmm15, %%xmm3			\n\t"
	    "phaddw	%%xmm3, %%xmm2			\n\t"
	    "paddw	%%xmm5, %%xmm2			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm2			\n\t"

//-----

            "prefetchnta 128(%1)			\n\t"
	    
	    "movdqa	%%xmm7, %%xmm5			\n\t"
	    "movdqa	%%xmm8, %%xmm6			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

            "movdqa	96(%1,%2,4), %%xmm9		\n\t"
            "movdqa	112(%1,%2,4), %%xmm10		\n\t"

	    "pshufb	%%xmm15, %%xmm7			\n\t"
	    "pshufb	%%xmm15, %%xmm8			\n\t"
	    "phaddw	%%xmm8, %%xmm7			\n\t"
	    "paddw	%%xmm5, %%xmm7			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm7			\n\t"
//------
	    
	    "movdqa	%%xmm9, %%xmm5			\n\t"
	    "movdqa	%%xmm10, %%xmm6			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm9			\n\t"
	    "pshufb	%%xmm15, %%xmm10		\n\t"
	    "phaddw	%%xmm10, %%xmm9			\n\t"
	    "paddw	%%xmm5, %%xmm9			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm9			\n\t"
//-----

	    "packuswb	%%xmm2, %%xmm0			\n\t"
	    "packuswb	%%xmm9, %%xmm7			\n\t"
            "movntps	%%xmm0, (%0,%2)			\n\t"
            "movntps	%%xmm7, 16(%0,%2)		\n\t"

            "add	$32, %2				\n\t"
            "jnz 	1b				\n\t"
            "pop 	%2				\n\t"

    :
    : "p" (dst2), "p" (src2), "r" (val), "p" (&shuf)
    );
	}
    MEASURE_STOP
    MEASURE_PRINT
}

void do9(unsigned char *dst, unsigned char *src) {
    int i;
    volatile int j;		// to prevent reduction of the loop
    unsigned char *end;

    MEASURE_DECLARE

    MEASURE_START
    end = dst + SIZE;
    
    ALGNW char shuf[64] ALGNL;
    char *shuf1 = shuf;
    char *shuf2 = shuf+16;
    char *shuf3 = shuf+32;
    uint16_t *mul = (uint16_t*)(shuf+48);

    memset(shuf, 0xFF, 48);
    
    for (i=0;i<16;i+=4) {
	shuf1[i] = i;
//	shuf1[i+1] = 0;
	shuf1[i+2] = i + 1;
//	shuf1[i+3] = 0;
    }

    for (i=0;i<8;i+=2) {
	shuf2[i] = 2 * i + 2;
//	shuf2[i+1] = 0;
	
	shuf3[i + 8] = 2 * i + 2;
//	shuf3[i + 9] = 0;
    }
    for (i=0;i<8;i++) {
	mul[i] = 0x2AAA;
    }

    long max = SIZE;
    long val = -max;
    char *src2 = src + 4 * SIZE;
    char *dst2 = dst + SIZE;
    
    for (j = 0; j < ITERATIONS; j++) {
        __asm__ __volatile__ (
            "prefetchnta (%1)				\n\t"
            "push 	%2				\n\t"
	    "movdqa	(%3), %%xmm15			\n\t"
	    "movdqa	16(%3), %%xmm14			\n\t"
	    "movdqa	32(%3), %%xmm13			\n\t"
	    "movdqa	48(%3), %%xmm12			\n\t"
        "1:						\n\t"
            "movdqa	(%1,%2,4), %%xmm0		\n\t"
            "movdqa	16(%1,%2,4), %%xmm1		\n\t"
        
	    "prefetchnta 64(%1)				\n\t"
	    
	    "movdqa	%%xmm0, %%xmm5			\n\t"
	    "pshufb	%%xmm15, %%xmm0			\n\t"

	    "movdqa	%%xmm1, %%xmm6			\n\t"
	    "pshufb	%%xmm15, %%xmm1			\n\t"

            "movdqa	32(%1,%2,4), %%xmm2		\n\t"
            "movdqa	48(%1,%2,4), %%xmm3		\n\t"

	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"


	    "phaddw	%%xmm1, %%xmm0			\n\t"
	    "paddw	%%xmm5, %%xmm0			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm0			\n\t"
//------
	    
	    "movdqa	%%xmm2, %%xmm5			\n\t"
	    "pshufb	%%xmm15, %%xmm2			\n\t"
	    "movdqa	%%xmm3, %%xmm6			\n\t"
	    "pshufb	%%xmm15, %%xmm3			\n\t"

            "movdqa	64(%1,%2,4), %%xmm7		\n\t"
            "movdqa	80(%1,%2,4), %%xmm8		\n\t"

	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "phaddw	%%xmm3, %%xmm2			\n\t"
	    "paddw	%%xmm5, %%xmm2			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm2			\n\t"

//-----

            "prefetchnta 128(%1)			\n\t"
	    
	    "movdqa	%%xmm7, %%xmm5			\n\t"
	    "pshufb	%%xmm15, %%xmm7			\n\t"
	    "movdqa	%%xmm8, %%xmm6			\n\t"
	    "pshufb	%%xmm15, %%xmm8			\n\t"

            "movdqa	96(%1,%2,4), %%xmm9		\n\t"
            "movdqa	112(%1,%2,4), %%xmm10		\n\t"

	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "phaddw	%%xmm8, %%xmm7			\n\t"
	    "paddw	%%xmm5, %%xmm7			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm7			\n\t"
//------
	    
	    "movdqa	%%xmm9, %%xmm5			\n\t"
	    "pshufb	%%xmm15, %%xmm9			\n\t"
	    "movdqa	%%xmm10, %%xmm6			\n\t"
	    "pshufb	%%xmm15, %%xmm10		\n\t"

	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "phaddw	%%xmm10, %%xmm9			\n\t"
	    "paddw	%%xmm5, %%xmm9			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm9			\n\t"
//-----

	    "packuswb	%%xmm2, %%xmm0			\n\t"
	    "packuswb	%%xmm9, %%xmm7			\n\t"
            "movntps	%%xmm0, (%0,%2)			\n\t"
            "movntps	%%xmm7, 16(%0,%2)		\n\t"

            "add	$32, %2				\n\t"
            "jnz 	1b				\n\t"
            "pop 	%2				\n\t"

    :
    : "p" (dst2), "p" (src2), "r" (val), "p" (&shuf)
    );
	}
    MEASURE_STOP
    MEASURE_PRINT
}

    // needs extra 32 bytes in the end of array
void do20(unsigned char *dst, unsigned char *src) {
    int i;
    volatile int j;		// to prevent reduction of the loop
    unsigned char *end;

    MEASURE_DECLARE

    MEASURE_START
    end = dst + SIZE;
    
    ALGNW char shuf[64] ALGNL;
    char *shuf1 = shuf;
    char *shuf2 = shuf+16;
    char *shuf3 = shuf+32;
    uint16_t *mul = (uint16_t*)(shuf+48);

    memset(shuf, 0xFF, 48);
    
    for (i=0;i<16;i+=4) {
	shuf1[i] = i;
//	shuf1[i+1] = 0;
	shuf1[i+2] = i + 1;
//	shuf1[i+3] = 0;
    }

    for (i=0;i<8;i+=2) {
	shuf2[i] = 2 * i + 2;
//	shuf2[i+1] = 0;
	
	shuf3[i + 8] = 2 * i + 2;
//	shuf3[i + 9] = 0;
    }
    for (i=0;i<8;i++) {
	mul[i] = 0x2AAA;
    }

    long max = SIZE;
    long val = -max;
    char *src2 = src + 4 * SIZE;
    char *dst2 = dst + SIZE;
    
    for (j = 0; j < ITERATIONS; j++) {
        __asm__ __volatile__ (
            "prefetchnta (%1)				\n\t"
            "push 	%2				\n\t"
	    "movdqa	(%3), %%xmm15			\n\t"
	    "movdqa	16(%3), %%xmm14			\n\t"
	    "movdqa	32(%3), %%xmm13			\n\t"
	    "movdqa	48(%3), %%xmm12			\n\t"

            "movdqa	(%1,%2,4), %%xmm0		\n\t"
            "movdqa	16(%1,%2,4), %%xmm1		\n\t"
        "1:						\n\t"
	    "prefetchnta 64(%1)				\n\t"
	    
	    "movdqa	%%xmm0, %%xmm5			\n\t"
	    "pshufb	%%xmm15, %%xmm0			\n\t"

	    "movdqa	%%xmm1, %%xmm6			\n\t"
	    "pshufb	%%xmm15, %%xmm1			\n\t"

            "movdqa	32(%1,%2,4), %%xmm2		\n\t"
            "movdqa	48(%1,%2,4), %%xmm3		\n\t"

	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"


	    "phaddw	%%xmm1, %%xmm0			\n\t"
	    "paddw	%%xmm5, %%xmm0			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm0			\n\t"
//------
	    
	    "movdqa	%%xmm2, %%xmm5			\n\t"
	    "pshufb	%%xmm15, %%xmm2			\n\t"
	    "movdqa	%%xmm3, %%xmm6			\n\t"
	    "pshufb	%%xmm15, %%xmm3			\n\t"

            "movdqa	64(%1,%2,4), %%xmm7		\n\t"
            "movdqa	80(%1,%2,4), %%xmm8		\n\t"

	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "phaddw	%%xmm3, %%xmm2			\n\t"
	    "paddw	%%xmm5, %%xmm2			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm2			\n\t"

//-----
	    "packuswb	%%xmm2, %%xmm0			\n\t"
            "movntps	%%xmm0, (%0,%2)			\n\t"

            "prefetchnta 128(%1)			\n\t"
	    
	    "movdqa	%%xmm7, %%xmm5			\n\t"
	    "pshufb	%%xmm15, %%xmm7			\n\t"
	    "movdqa	%%xmm8, %%xmm6			\n\t"
	    "pshufb	%%xmm15, %%xmm8			\n\t"

            "movdqa	96(%1,%2,4), %%xmm9		\n\t"
            "movdqa	112(%1,%2,4), %%xmm10		\n\t"

	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "phaddw	%%xmm8, %%xmm7			\n\t"
	    "paddw	%%xmm5, %%xmm7			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm7			\n\t"
//------
	    
	    "movdqa	%%xmm9, %%xmm5			\n\t"
	    "pshufb	%%xmm15, %%xmm9			\n\t"
	    "movdqa	%%xmm10, %%xmm6			\n\t"
	    "pshufb	%%xmm15, %%xmm10		\n\t"

	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

            "movdqa	128(%1,%2,4), %%xmm0		\n\t"
            "movdqa	144(%1,%2,4), %%xmm1		\n\t"

	    "phaddw	%%xmm10, %%xmm9			\n\t"
	    "paddw	%%xmm5, %%xmm9			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm9			\n\t"
//-----

	    "packuswb	%%xmm9, %%xmm7			\n\t"
            "movntps	%%xmm7, 16(%0,%2)		\n\t"

            "add	$32, %2				\n\t"
            "jnz 	1b				\n\t"
            "pop 	%2				\n\t"

    :
    : "p" (dst2), "p" (src2), "r" (val), "p" (&shuf)
    );
	}
    MEASURE_STOP
    MEASURE_PRINT
}

// and another 32 before
void do21(unsigned char *dst, unsigned char *src) {
    int i;
    volatile int j;		// to prevent reduction of the loop
    unsigned char *end;

    MEASURE_DECLARE

    MEASURE_START
    end = dst + SIZE;
    
    ALGNW char shuf[64] ALGNL;
    char *shuf1 = shuf;
    char *shuf2 = shuf+16;
    char *shuf3 = shuf+32;
    uint16_t *mul = (uint16_t*)(shuf+48);

    memset(shuf, 0xFF, 48);
    
    for (i=0;i<16;i+=4) {
	shuf1[i] = i;
//	shuf1[i+1] = 0;
	shuf1[i+2] = i + 1;
//	shuf1[i+3] = 0;
    }

    for (i=0;i<8;i+=2) {
	shuf2[i] = 2 * i + 2;
//	shuf2[i+1] = 0;
	
	shuf3[i + 8] = 2 * i + 2;
//	shuf3[i + 9] = 0;
    }
    for (i=0;i<8;i++) {
	mul[i] = 0x2AAA;
    }

    long max = SIZE;
    long val = -max;
    char *src2 = src + 4 * SIZE;
    char *dst2 = dst + SIZE;
    
    for (j = 0; j < ITERATIONS; j++) {
        __asm__ __volatile__ (
            "prefetchnta (%1)				\n\t"
            "push 	%2				\n\t"
	    "movdqa	(%3), %%xmm15			\n\t"
	    "movdqa	16(%3), %%xmm14			\n\t"
	    "movdqa	32(%3), %%xmm13			\n\t"
	    "movdqa	48(%3), %%xmm12			\n\t"

            "movdqa	(%1,%2,4), %%xmm0		\n\t"
            "movdqa	16(%1,%2,4), %%xmm1		\n\t"
        "1:						\n\t"
	    "prefetchnta 64(%1)				\n\t"
	    
	    "movdqa	%%xmm0, %%xmm5			\n\t"
	    "pshufb	%%xmm15, %%xmm0			\n\t"

	    "packuswb	%%xmm9, %%xmm7			\n\t"
            "movntps	%%xmm7, -16(%0,%2)		\n\t"

	    "movdqa	%%xmm1, %%xmm6			\n\t"
	    "pshufb	%%xmm15, %%xmm1			\n\t"

            "movdqa	32(%1,%2,4), %%xmm2		\n\t"
            "movdqa	48(%1,%2,4), %%xmm3		\n\t"

	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"


	    "phaddw	%%xmm1, %%xmm0			\n\t"
	    "paddw	%%xmm5, %%xmm0			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm0			\n\t"
//------
	    
	    "movdqa	%%xmm2, %%xmm5			\n\t"
	    "pshufb	%%xmm15, %%xmm2			\n\t"
	    "movdqa	%%xmm3, %%xmm6			\n\t"
	    "pshufb	%%xmm15, %%xmm3			\n\t"

            "movdqa	64(%1,%2,4), %%xmm7		\n\t"
            "movdqa	80(%1,%2,4), %%xmm8		\n\t"

	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "phaddw	%%xmm3, %%xmm2			\n\t"
	    "paddw	%%xmm5, %%xmm2			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm2			\n\t"

//-----
            "prefetchnta 128(%1)			\n\t"
	    
	    "movdqa	%%xmm7, %%xmm5			\n\t"
	    "pshufb	%%xmm15, %%xmm7			\n\t"

	    "packuswb	%%xmm2, %%xmm0			\n\t"
            "movntps	%%xmm0, (%0,%2)			\n\t"

	    "movdqa	%%xmm8, %%xmm6			\n\t"
	    "pshufb	%%xmm15, %%xmm8			\n\t"

            "movdqa	96(%1,%2,4), %%xmm9		\n\t"
            "movdqa	112(%1,%2,4), %%xmm10		\n\t"

	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "phaddw	%%xmm8, %%xmm7			\n\t"
	    "paddw	%%xmm5, %%xmm7			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm7			\n\t"
//------
	    
	    "movdqa	%%xmm9, %%xmm5			\n\t"
	    "pshufb	%%xmm15, %%xmm9			\n\t"
	    "movdqa	%%xmm10, %%xmm6			\n\t"
	    "pshufb	%%xmm15, %%xmm10		\n\t"

	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

            "movdqa	128(%1,%2,4), %%xmm0		\n\t"
            "movdqa	144(%1,%2,4), %%xmm1		\n\t"

	    "phaddw	%%xmm10, %%xmm9			\n\t"
	    "paddw	%%xmm5, %%xmm9			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm9			\n\t"
//-----

            "add	$32, %2				\n\t"
            "jnz 	1b				\n\t"

	    "packuswb	%%xmm9, %%xmm7			\n\t"
            "movntps	%%xmm7, -16(%0,%2)		\n\t"

            "pop 	%2				\n\t"

    :
    : "p" (dst2), "p" (src2), "r" (val), "p" (&shuf)
    );
	}
    MEASURE_STOP
    MEASURE_PRINT
}

void do22(unsigned char *dst, unsigned char *src) {
    int i;
    volatile int j;		// to prevent reduction of the loop
    unsigned char *end;

    MEASURE_DECLARE

    MEASURE_START
    end = dst + SIZE;
    
    ALGNW char shuf[64] ALGNL;
    char *shuf1 = shuf;
    char *shuf2 = shuf+16;
    char *shuf3 = shuf+32;
    uint16_t *mul = (uint16_t*)(shuf+48);

    memset(shuf, 0xFF, 48);
    
    for (i=0;i<16;i+=4) {
	shuf1[i] = i;
//	shuf1[i+1] = 0;
	shuf1[i+2] = i + 1;
//	shuf1[i+3] = 0;
    }

    for (i=0;i<8;i+=2) {
	shuf2[i] = 2 * i + 2;
//	shuf2[i+1] = 0;
	
	shuf3[i + 8] = 2 * i + 2;
//	shuf3[i + 9] = 0;
    }
    for (i=0;i<8;i++) {
	mul[i] = 0x2AAA;
    }

    long max = SIZE;
    long val = -max;
    char *src2 = src + 4 * SIZE;
    char *dst2 = dst + SIZE;
    
    for (j = 0; j < ITERATIONS; j++) {
        __asm__ __volatile__ (
            "prefetchnta (%1)				\n\t"
            "push 	%2				\n\t"
	    "movdqa	(%3), %%xmm15			\n\t"
	    "movdqa	16(%3), %%xmm14			\n\t"
	    "movdqa	32(%3), %%xmm13			\n\t"
	    "movdqa	48(%3), %%xmm12			\n\t"

            "movdqa	(%1,%2,4), %%xmm0		\n\t"
            "movdqa	16(%1,%2,4), %%xmm1		\n\t"
        "1:						\n\t"
	    "prefetchnta 64(%1)				\n\t"
	    
	    "movdqa	%%xmm0, %%xmm5			\n\t"
	    "pshufb	%%xmm15, %%xmm0			\n\t"

	    "pmulhrsw	%%xmm12, %%xmm9			\n\t"

	    "movdqa	%%xmm1, %%xmm6			\n\t"
	    "pshufb	%%xmm15, %%xmm1			\n\t"

	    "packuswb	%%xmm9, %%xmm7			\n\t"
            "movntps	%%xmm7, -16(%0,%2)		\n\t"

            "movdqa	32(%1,%2,4), %%xmm2		\n\t"
            "movdqa	48(%1,%2,4), %%xmm3		\n\t"

	    "phaddw	%%xmm1, %%xmm0			\n\t"

	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

//------

	    "movdqa	%%xmm3, %%xmm6			\n\t"
	    "pshufb	%%xmm15, %%xmm3			\n\t"

	    "paddw	%%xmm5, %%xmm0			\n\t"
	    
	    "movdqa	%%xmm2, %%xmm5			\n\t"
	    "pshufb	%%xmm15, %%xmm2			\n\t"

	    "pmulhrsw	%%xmm12, %%xmm0			\n\t"

            "movdqa	64(%1,%2,4), %%xmm7		\n\t"
            "movdqa	80(%1,%2,4), %%xmm8		\n\t"

	    "phaddw	%%xmm3, %%xmm2			\n\t"

	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"

	    "por	%%xmm6, %%xmm5			\n\t"
	    "paddw	%%xmm5, %%xmm2			\n\t"
//-----
            "prefetchnta 128(%1)			\n\t"
	    
	    "movdqa	%%xmm7, %%xmm11			\n\t"//5
	    "pshufb	%%xmm15, %%xmm7			\n\t"

	    "pmulhrsw	%%xmm12, %%xmm2			\n\t"
	    "movdqa	%%xmm8, %%xmm6			\n\t"
	    "pshufb	%%xmm15, %%xmm8			\n\t"

	    "packuswb	%%xmm2, %%xmm0			\n\t"
            "movntps	%%xmm0, (%0,%2)			\n\t"

            "movdqa	96(%1,%2,4), %%xmm9		\n\t"
            "movdqa	112(%1,%2,4), %%xmm10		\n\t"

	    "phaddw	%%xmm8, %%xmm7			\n\t"

	    "pshufb	%%xmm14, %%xmm11		\n\t"//5
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm11			\n\t"//5

//------

	    "movdqa	%%xmm10, %%xmm6			\n\t"
	    "pshufb	%%xmm15, %%xmm10		\n\t"

	    "paddw	%%xmm11, %%xmm7			\n\t"//5

	    "movdqa	%%xmm9, %%xmm5			\n\t"
	    "pshufb	%%xmm15, %%xmm9			\n\t"

	    "pmulhrsw	%%xmm12, %%xmm7			\n\t"

	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"


            "movdqa	128(%1,%2,4), %%xmm0		\n\t"
	    "phaddw	%%xmm10, %%xmm9			\n\t"
            "movdqa	144(%1,%2,4), %%xmm1		\n\t"

	    "paddw	%%xmm5, %%xmm9			\n\t"
//-----

            "add	$32, %2				\n\t"
            "jnz 	1b				\n\t"

	    "pmulhrsw	%%xmm12, %%xmm9			\n\t"
	    "packuswb	%%xmm9, %%xmm7			\n\t"
            "movntps	%%xmm7, -16(%0,%2)		\n\t"

            "pop 	%2				\n\t"

    :
    : "p" (dst2), "p" (src2), "r" (val), "p" (&shuf)
    );
	}
    MEASURE_STOP
    MEASURE_PRINT
}

void do23(unsigned char *dst, unsigned char *src) {
    int i;
    volatile int j;		// to prevent reduction of the loop
    unsigned char *end;

    MEASURE_DECLARE

    MEASURE_START
    end = dst + SIZE;
    
    ALGNW char shuf[64] ALGNL;
    char *shuf1 = shuf;
    char *shuf2 = shuf+16;
    char *shuf3 = shuf+32;
    uint16_t *mul = (uint16_t*)(shuf+48);

    memset(shuf, 0xFF, 48);
    
    for (i=0;i<16;i+=4) {
	shuf1[i] = i;
//	shuf1[i+1] = 0;
	shuf1[i+2] = i + 1;
//	shuf1[i+3] = 0;
    }

    for (i=0;i<8;i+=2) {
	shuf2[i] = 2 * i + 2;
//	shuf2[i+1] = 0;
	
	shuf3[i + 8] = 2 * i + 2;
//	shuf3[i + 9] = 0;
    }
    for (i=0;i<8;i++) {
	mul[i] = 0x2AAA;
    }

    long max = SIZE;
    long val = -max;
    char *src2 = src + 4 * SIZE;
    char *dst2 = dst + SIZE;
    
    for (j = 0; j < ITERATIONS; j++) {
        __asm__ __volatile__ (
            "prefetchnta (%1,%2,4)			\n\t"
            "push 	%2				\n\t"
	    "movdqa	(%3), %%xmm15			\n\t"
	    "movdqa	16(%3), %%xmm14			\n\t"
	    "movdqa	32(%3), %%xmm13			\n\t"
	    "movdqa	48(%3), %%xmm12			\n\t"

            "movdqa	(%1,%2,4), %%xmm0		\n\t"
            "movdqa	16(%1,%2,4), %%xmm1		\n\t"
        "1:						\n\t"
	    "prefetchnta 64(%1,%2,4)			\n\t"
	    
	    "movdqa	%%xmm0, %%xmm5			\n\t"
	    "pshufb	%%xmm15, %%xmm0			\n\t"

	    "pmulhrsw	%%xmm12, %%xmm9			\n\t"

	    "movdqa	%%xmm1, %%xmm6			\n\t"
	    "pshufb	%%xmm15, %%xmm1			\n\t"

	    "packuswb	%%xmm9, %%xmm7			\n\t"
            "movntps	%%xmm7, -16(%0,%2)		\n\t"

            "movdqa	32(%1,%2,4), %%xmm2		\n\t"
            "movdqa	48(%1,%2,4), %%xmm3		\n\t"

	    "phaddw	%%xmm1, %%xmm0			\n\t"

	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

//------

	    "movdqa	%%xmm3, %%xmm6			\n\t"
	    "pshufb	%%xmm15, %%xmm3			\n\t"

	    "paddw	%%xmm5, %%xmm0			\n\t"
	    
	    "movdqa	%%xmm2, %%xmm5			\n\t"
	    "pshufb	%%xmm15, %%xmm2			\n\t"

	    "pmulhrsw	%%xmm12, %%xmm0			\n\t"

            "movdqa	64(%1,%2,4), %%xmm7		\n\t"
            "movdqa	80(%1,%2,4), %%xmm8		\n\t"

	    "phaddw	%%xmm3, %%xmm2			\n\t"

	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"

//-----
            "prefetchnta 128(%1,%2,4)			\n\t"
	    
	    "movdqa	%%xmm7, %%xmm11			\n\t"//5
	    "por	%%xmm6, %%xmm5			\n\t"
	    "pshufb	%%xmm15, %%xmm7			\n\t"
	    "paddw	%%xmm5, %%xmm2			\n\t"

	    "movdqa	%%xmm8, %%xmm6			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm2			\n\t"
	    "pshufb	%%xmm15, %%xmm8			\n\t"

            "movdqa	96(%1,%2,4), %%xmm9		\n\t"
	    "packuswb	%%xmm2, %%xmm0			\n\t"
            "movntps	%%xmm0, (%0,%2)			\n\t"
            "movdqa	112(%1,%2,4), %%xmm10		\n\t"

	    "phaddw	%%xmm8, %%xmm7			\n\t"

	    "pshufb	%%xmm14, %%xmm11		\n\t"//5
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm11			\n\t"//5

//------

	    "movdqa	%%xmm10, %%xmm6			\n\t"
	    "pshufb	%%xmm15, %%xmm10		\n\t"

	    "paddw	%%xmm11, %%xmm7			\n\t"//5

	    "movdqa	%%xmm9, %%xmm5			\n\t"
	    "pshufb	%%xmm15, %%xmm9			\n\t"

	    "pmulhrsw	%%xmm12, %%xmm7			\n\t"

	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"


            "movdqa	128(%1,%2,4), %%xmm0		\n\t"
	    "phaddw	%%xmm10, %%xmm9			\n\t"
            "movdqa	144(%1,%2,4), %%xmm1		\n\t"

	    "paddw	%%xmm5, %%xmm9			\n\t"
//-----

            "add	$32, %2				\n\t"
            "jnz 	1b				\n\t"

	    "pmulhrsw	%%xmm12, %%xmm9			\n\t"
	    "packuswb	%%xmm9, %%xmm7			\n\t"
            "movntps	%%xmm7, -16(%0,%2)		\n\t"

            "pop 	%2				\n\t"

    :
    : "p" (dst2), "p" (src2), "r" (val), "p" (&shuf)
    );
	}
    MEASURE_STOP
    MEASURE_PRINT
}




typedef struct TTHREADParams THREADParams;
struct TTHREADParams {
    char *src;
    char *dst;
    
    int start;
    int size;

    char *shuf;
    
    int iters;
};


void *compute(THREADParams *ctx) {
    int i;
    long val = -ctx->size;
    unsigned char *src2 = ctx->src + 4 * (ctx->start + ctx->size);
    unsigned char *dst2 = ctx->dst + (ctx->start + ctx->size);

    char *shuf = ctx->shuf;
    char *shuf1 = shuf;
    char *shuf2 = shuf+16;
    char *shuf3 = shuf+32;
    uint16_t *mul = (uint16_t*)(shuf+48);

    int iters = ctx->iters;
    
    for (i = 0; i < iters; i++) {
	__asm__ __volatile__ (
            "prefetchnta (%1)				\n\t"
            "push 	%2				\n\t"
	    "movdqa	(%3), %%xmm15			\n\t"
	    "movdqa	16(%3), %%xmm14			\n\t"
	    "movdqa	32(%3), %%xmm13			\n\t"
	    "movdqa	48(%3), %%xmm12			\n\t"
            "1:						\n\t"
            "movdqa	(%1,%2,4), %%xmm0		\n\t"
            "movdqa	16(%1,%2,4), %%xmm1		\n\t"
        
	    "prefetchnta 64(%1)				\n\t"
	    
	    "movdqa	%%xmm0, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm1, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm0			\n\t"
	    "pshufb	%%xmm15, %%xmm1			\n\t"
	    "phaddw	%%xmm1, %%xmm0			\n\t"
	    "paddw	%%xmm5, %%xmm0			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm0			\n\t"
//------
            "movdqa	32(%1,%2,4), %%xmm1		\n\t"
            "movdqa	48(%1,%2,4), %%xmm2		\n\t"
	    
	    "movdqa	%%xmm1, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm2, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm1			\n\t"
	    "pshufb	%%xmm15, %%xmm2			\n\t"
	    "phaddw	%%xmm2, %%xmm1			\n\t"
	    "paddw	%%xmm5, %%xmm1			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm1			\n\t"

//-----
            "movdqa	64(%1,%2,4), %%xmm2		\n\t"
            "movdqa	80(%1,%2,4), %%xmm3		\n\t"

            "prefetchnta 128(%1)			\n\t"
	    
	    "movdqa	%%xmm2, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm3, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm2			\n\t"
	    "pshufb	%%xmm15, %%xmm3			\n\t"
	    "phaddw	%%xmm3, %%xmm2			\n\t"
	    "paddw	%%xmm5, %%xmm2			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm2			\n\t"
//------
            "movdqa	96(%1,%2,4), %%xmm3		\n\t"
            "movdqa	112(%1,%2,4), %%xmm4		\n\t"
	    
	    "movdqa	%%xmm3, %%xmm5			\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
	    "movdqa	%%xmm4, %%xmm6			\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm3			\n\t"
	    "pshufb	%%xmm15, %%xmm4			\n\t"
	    "phaddw	%%xmm4, %%xmm3			\n\t"
	    "paddw	%%xmm5, %%xmm3			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm3			\n\t"
//-----

	    "packuswb	%%xmm1, %%xmm0			\n\t"
	    "packuswb	%%xmm3, %%xmm2			\n\t"
            "movntps	%%xmm0, (%0,%2)			\n\t"
            "movntps	%%xmm2, 16(%0,%2)		\n\t"

            "add	$32, %2				\n\t"
            "jnz 	1b				\n\t"
            "pop 	%2				\n\t"

	:
	: "p" (dst2), "p" (src2), "r" (val), "p" (shuf)
	);
    }
    
    pthread_exit(NULL);
}


void do10(unsigned char *dst, unsigned char *src) {
    int i;
    int err;
    volatile int j;		// to prevent reduction of the loop

    pthread_t threads[num_procs];
    THREADParams params[num_procs];

    MEASURE_DECLARE

    MEASURE_START
    
    ALGNW char shuf[64] ALGNL;
    char *shuf1 = shuf;
    char *shuf2 = shuf+16;
    char *shuf3 = shuf+32;
    uint16_t *mul = (uint16_t*)(shuf+48);

    memset(shuf, 0xFF, 48);
    
    for (i=0;i<16;i+=4) {
	shuf1[i] = i;
	shuf1[i+2] = i + 1;
    }

    for (i=0;i<8;i+=2) {
	shuf2[i] = 2 * i + 2;
	shuf3[i + 8] = 2 * i + 2;
    }

    for (i=0;i<8;i++) {
	mul[i] = 0x2AAA;
    }

    int per_proc = SIZE / num_procs;
    per_proc = calc_alloc(per_proc, BLOCK);
    
    int num_threads = 0;
    for (i=0;i<SIZE;i+=per_proc) {
	params[num_threads].src = src;
	params[num_threads].dst = dst;
	params[num_threads].start = i;
        params[num_threads].size = min2(per_proc, SIZE - i);
	params[num_threads].shuf = shuf;
	params[num_threads].iters = 1;

	num_threads++;
    }

//    printf("%i = %i * %i\n", SIZE, num_threads, per_proc);

    for (j = 0; j < ITERATIONS; j++) {
	for (i = 0; i < num_threads; i++) {
	    pthread_create(&threads[i], NULL, (void *(*)(void*))compute, &params[i]);
	}

	for (i = 0; i < num_threads; i++) {
	    pthread_join(threads[i], NULL);
	}
    }
    
    MEASURE_STOP
    MEASURE_PRINT
    
}


void do11(unsigned char *dst, unsigned char *src, int procs) {
    int i;
    int err;
    volatile int j;		// to prevent reduction of the loop

    pthread_t threads[procs];
    
    THREADParams params[procs];

    MEASURE_DECLARE

    MEASURE_START
    
    ALGNW char shuf[64] ALGNL;
    char *shuf1 = shuf;
    char *shuf2 = shuf+16;
    char *shuf3 = shuf+32;
    uint16_t *mul = (uint16_t*)(shuf+48);

    memset(shuf, 0xFF, 48);
    
    for (i=0;i<16;i+=4) {
	shuf1[i] = i;
	shuf1[i+2] = i + 1;
    }

    for (i=0;i<8;i+=2) {
	shuf2[i] = 2 * i + 2;
	shuf3[i + 8] = 2 * i + 2;
    }

    for (i=0;i<8;i++) {
	mul[i] = 0x2AAA;
    }

    int per_proc = SIZE / procs;
    per_proc = calc_alloc(per_proc, BLOCK);
    
    int num_threads = 0;
    for (i=0;i<SIZE;i+=per_proc) {
	params[num_threads].src = src;
	params[num_threads].dst = dst;
	params[num_threads].start = i;
        params[num_threads].size = min2(per_proc, SIZE - i);
	params[num_threads].shuf = shuf;
	params[num_threads].iters = ITERATIONS;

	num_threads++;
    }
    
    for (i = 0; i < num_threads; i++) {
	pthread_create(&threads[i], NULL, (void *(*)(void*))compute, &params[i]);
    }

    for (i = 0; i < num_threads; i++) {
        pthread_join(threads[i], NULL);
    }
    
    MEASURE_STOP
    MEASURE_PRINT_PROCS
}




void do12(unsigned char *dst, unsigned char *src) {
    int i;
    volatile int j;		// to prevent reduction of the loop
    unsigned char *end;

    MEASURE_DECLARE

    MEASURE_START
    end = dst + SIZE;
    
    ALGNW char shuf[64] ALGNL;
    char *shuf1 = shuf;
    char *shuf2 = shuf+16;
    char *shuf3 = shuf+32;
    uint16_t *mul = (uint16_t*)(shuf+48);

    memset(shuf, 0xFF, 48);
    
    for (i=0;i<16;i+=4) {
	shuf1[i] = i;
//	shuf1[i+1] = 0;
	shuf1[i+2] = i + 1;
//	shuf1[i+3] = 0;
    }

    for (i=0;i<8;i+=2) {
	shuf2[i] = 2 * i + 2;
//	shuf2[i+1] = 0;
	
	shuf3[i + 8] = 2 * i + 2;
//	shuf3[i + 9] = 0;
    }

    for (i=0;i<8;i++) {
	mul[i] = 0x2AAA;
    }


//    int per_proc = SIZE / num_procs;
    int blocks = calc_blocks(SIZE, BLOCK);


    for (j = 0; j < ITERATIONS; j++) {
	#pragma omp parallel for
	for (i = 0; i < blocks; i++) {
	    long val = -BLOCK;
	    char *src2 = src + 4 * (i + 1)*BLOCK;
	    char *dst2 = dst + (i + 1) * BLOCK;
	    
        __asm__ __volatile__ (
            "prefetchnta (%1)				\n\t"
            "push 	%2				\n\t"
	    "movdqa	(%3), %%xmm15			\n\t"
	    "movdqa	16(%3), %%xmm14			\n\t"
	    "movdqa	32(%3), %%xmm13			\n\t"
	    "movdqa	48(%3), %%xmm12			\n\t"
            "1:						\n\t"
            "movdqa	(%1,%2,4), %%xmm0		\n\t"
            "movdqa	16(%1,%2,4), %%xmm1		\n\t"
        
	    "prefetchnta 64(%1)				\n\t"
	    
//	    "movdqa	%%xmm0, %%xmm5			\n\t"
	    "pshufd	$0xE4, %%xmm0, %%xmm5		\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
//	    "movdqa	%%xmm1, %%xmm6			\n\t"
	    "pshufd	$0xE4, %%xmm1, %%xmm6		\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm0			\n\t"
	    "pshufb	%%xmm15, %%xmm1			\n\t"
	    "phaddw	%%xmm1, %%xmm0			\n\t"
	    "paddw	%%xmm5, %%xmm0			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm0			\n\t"
//------
            "movdqa	32(%1,%2,4), %%xmm1		\n\t"
            "movdqa	48(%1,%2,4), %%xmm2		\n\t"
	    
//	    "movdqa	%%xmm1, %%xmm5			\n\t"
	    "pshufd	$0xE4, %%xmm1, %%xmm5		\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
//	    "movdqa	%%xmm2, %%xmm6			\n\t"
	    "pshufd	$0xE4, %%xmm2, %%xmm6		\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm1			\n\t"
	    "pshufb	%%xmm15, %%xmm2			\n\t"
	    "phaddw	%%xmm2, %%xmm1			\n\t"
	    "paddw	%%xmm5, %%xmm1			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm1			\n\t"

//-----
            "movdqa	64(%1,%2,4), %%xmm2		\n\t"
            "movdqa	80(%1,%2,4), %%xmm3		\n\t"

            "prefetchnta 128(%1)			\n\t"
	    
//	    "movdqa	%%xmm2, %%xmm5			\n\t"
	    "pshufd	$0xE4, %%xmm2, %%xmm5		\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
//	    "movdqa	%%xmm3, %%xmm6			\n\t"
	    "pshufd	$0xE4, %%xmm3, %%xmm6		\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm2			\n\t"
	    "pshufb	%%xmm15, %%xmm3			\n\t"
	    "phaddw	%%xmm3, %%xmm2			\n\t"
	    "paddw	%%xmm5, %%xmm2			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm2			\n\t"
//------
            "movdqa	96(%1,%2,4), %%xmm3		\n\t"
            "movdqa	112(%1,%2,4), %%xmm4		\n\t"
	    
//	    "movdqa	%%xmm3, %%xmm5			\n\t"
	    "pshufd	$0xE4, %%xmm3, %%xmm5		\n\t"
	    "pshufb	%%xmm14, %%xmm5			\n\t"
//	    "movdqa	%%xmm4, %%xmm6			\n\t"
	    "pshufd	$0xE4, %%xmm4, %%xmm6		\n\t"
	    "pshufb	%%xmm13, %%xmm6			\n\t"
	    "por	%%xmm6, %%xmm5			\n\t"

	    "pshufb	%%xmm15, %%xmm3			\n\t"
	    "pshufb	%%xmm15, %%xmm4			\n\t"
	    "phaddw	%%xmm4, %%xmm3			\n\t"
	    "paddw	%%xmm5, %%xmm3			\n\t"
	    "pmulhrsw	%%xmm12, %%xmm3			\n\t"
//-----

	    "packuswb	%%xmm1, %%xmm0			\n\t"
	    "packuswb	%%xmm3, %%xmm2			\n\t"
            "movntps	%%xmm0, (%0,%2)			\n\t"
            "movntps	%%xmm2, 16(%0,%2)		\n\t"

            "add	$32, %2				\n\t"
            "jnz 	1b				\n\t"
            "pop 	%2				\n\t"

    :
    : "p" (dst2), "p" (src2), "r" (val), "p" (&shuf)
    );
	}
    }

    MEASURE_STOP
    MEASURE_PRINT
}









#define RESET \
 for (i=0;i<SIZE;i++) dst[i] = 0;

#define VALIDATE(test) \
 for (i=0;i<SIZE;i++) \
    if (dst[i] != gold[i]) { \
	printf("TEST %i: Error in position: %i, %i (test) != %i (gold) [%i,%i,%i]\n", test, i, dst[i], gold[i], src[4*i], src[4*i+1], src[4*i+2]); \
	exit(1); \
    }

#define TEST(i) \
 RESET \
 do##i(dst, src); \
 VALIDATE(i) \

#define TEST_PROCS(i, max_proc) \
 RESET \
 for (int j = 1; j <= max_proc; j++) { \
    do##i(dst, src, j); \
 } \
 VALIDATE(i) \

 
int main() {
    int i;
    unsigned char *real_dst, *real_src;
    unsigned char *dst;//, *gold;
    unsigned char *src;

    gold = (unsigned char*)malloc((SIZE+128)*sizeof(char));
//    dst = (unsigned char*)malloc((SIZE+128)*sizeof(char));
//    src = (unsigned char*)malloc(4*(SIZE+128)*sizeof(char));
    posix_memalign((void**)&real_dst, 16, (SIZE+256)*sizeof(char));
    posix_memalign((void**)&real_src, 16, 4*(SIZE+256)*sizeof(char));
    
    dst = real_dst + 128;
    src = real_src + 128;

    struct timeval tm;
    gettimeofday(&tm, NULL);
    srandom(tm.tv_usec);
    for (i=0;i<SIZE;i++) {
/*
	src[4*i] = (i + 1)%254;
	src[4*i+1] = src[4*i] - 1;
	src[4*i+2] = src[4*i] + 1;
	src[4*i+3] = 255;
*/

	int j = i%766;
	src[4*i] =min2(255, j); j -= src[4*i];
	src[4*i+1] = min2(255, j); j -= src[4*i+1];
	src[4*i+2] = j;
	src[4*i+3] = 255;

/*
	src[4*i] = random()%256;
	src[4*i+1] = random()%256;
	src[4*i+2] = random()%256;
	src[4*i+3] = 255;
*/
    }

//    do_gold(gold, src);

//    TEST(1)
    do6(gold, src);
//    TEST(2)
//    TEST(3)
    TEST(4)
//    TEST(5)
    
    TEST(6)
//    TEST(7)
//    TEST(8)
    TEST(9)
    TEST(20)
//    TEST(21)
    TEST(22)
    TEST(23)

/*
    //do6(gold, src);

    num_procs = omp_get_num_procs();
    printf("Starting OpenMP tests, number of processors: %d\n", num_procs);

    TEST(10)

    TEST_PROCS(11, num_procs);

    TEST(12)
*/

    TEST_PROCS(11, 2);

    free(real_src);
    free(real_dst);
    free(gold);
}
