#define _GNU_SOURCE
#include <stdio.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <string.h>
#include <smmintrin.h>
#include <strings.h>
#include <pthread.h>
#include <sched.h>
#include <errno.h>


#include "lib/memcpy.c"
//#define USE_CUDA

#ifdef USE_CUDA
# include <cuda.h>
# include <cuda_runtime.h>
#endif /* USE_CUDA */


#define THREADS 0
//#define THREADS 2
//#define MULTIMODE
#define ITERS 250l
#define SIZE (64 * 1024 * 1024l)



void fast_memcpy_gcc(void *dst, void *src, size_t size) {
    __builtin_memcpy(dst, src, size);
}

void fast_memcpy_glibc(void *dst, void *src, size_t size) {
    memcpy(dst, src, size);
}

void fast_memcpy_glibc_blocked(void *dst, void *src, size_t size) {
    int i,j,k;
    
    for (i = 0; i < size; i += 4 * 4096) {
	for (k = 0; k < 4096; k += 512) {

	    for (j = 0; j < 4; j++) {
		memcpy(dst + i + j * 4096 + k, src + i + j * 4096 + k, 512);
	    }
	}
    }
}


void fast_memcpy_bcopy(void *dst, void *src, size_t size) {
    bcopy(dst, src, size);
}

void fast_memcpy_vik(void *dst, void *src, size_t size) {
    fast_memcpy(dst, src, size);
}

void fast_memcpy_intrinsic(void *dst, void *src, size_t size) {
    char *dst_end=dst+size;
    while(dst!=dst_end){
        __m128i res = _mm_stream_load_si128((__m128i *)src);
        *((__m128i *)dst)=res;
        src+=16;
        dst+=16;
    }
}

#ifndef __INTEL_COMPILER
void fast_memcpy_repmovs(void *dst, void *src, size_t size) {
    __asm__ __volatile__ (
        "mov         %0, %%rdi                  \n\t"
        "mov         %1, %%rsi                  \n\t"
        "mov         %2, %%rcx                  \n\t"

        "shr         $3, %%rcx                  \n\t"
        "rep         movsq                      \n\t"

//        "shr         $2, %%rcx                  \n\t"
//        "rep         movsl                      \n\t"
//        "rep         movsb                      \n\t"
    :
    : "p" (dst), "p" (src), "r" (size)
    : "%esi", "%edi", "%ecx"
    );
}


void fast_memcpy_dqa_ntps(void *dst, void *src, size_t size) {
    size_t sse_size = (size / 128);

    __asm__ __volatile__ (
            "push 	%2			\n\t"
            "mov        $0, %%rax		\n\t"

            "1:					\n\t"
            "prefetchnta 128(%0,%%rax)		\n\t"
            "prefetchnta 160(%0,%%rax)		\n\t"
            "prefetchnta 192(%0,%%rax)		\n\t"
            "prefetchnta 224(%0,%%rax)		\n\t"


            "movdqa 	(%0,%%rax), %%xmm0	\n\t"
            "movdqa 	16(%0,%%rax), %%xmm1	\n\t"
            "movdqa 	32(%0,%%rax), %%xmm2	\n\t"
            "movdqa 	48(%0,%%rax), %%xmm3	\n\t"
            "movdqa 	64(%0,%%rax), %%xmm4	\n\t"
            "movdqa 	80(%0,%%rax), %%xmm5	\n\t"
            "movdqa 	96(%0,%%rax), %%xmm6	\n\t"
            "movdqa 	112(%0,%%rax), %%xmm7	\n\t"

            "movntps	%%xmm0, (%1,%%rax)	\n\t"
            "movntps	%%xmm1, 16(%1,%%rax)	\n\t"
            "movntps	%%xmm2, 32(%1,%%rax)	\n\t"
            "movntps	%%xmm3, 48(%1,%%rax)	\n\t"
            "movntps	%%xmm4, 64(%1,%%rax)	\n\t"
            "movntps	%%xmm5, 80(%1,%%rax)	\n\t"
            "movntps	%%xmm6, 96(%1,%%rax)	\n\t"
            "movntps	%%xmm7, 112(%1,%%rax)	\n\t"

            "add	$128, %%rax		\n\t"
            "dec	%2			\n\t"
            "jnz 	1b			\n\t"
            "pop 	%2			\n\t"

            "sfence"
    :
    : "p" (dst), "p" (src), "r" (sse_size)
    : "%rax"
        );
}

void fast_memcpy_ntdqa_ntps(void *dst, void *src, size_t size) {
    size_t sse_size = (size / 128);

    __asm__ __volatile__ (
            "push 	%2			\n\t"
            "mov        $0, %%rax		\n\t"

            "1:					\n\t"
            "prefetchnta 128(%0,%%rax)		\n\t"
            "prefetchnta 160(%0,%%rax)		\n\t"
            "prefetchnta 192(%0,%%rax)		\n\t"
            "prefetchnta 224(%0,%%rax)		\n\t"


            "movntdqa 	(%0,%%rax), %%xmm0	\n\t"
            "movntdqa 	16(%0,%%rax), %%xmm1	\n\t"
            "movntdqa 	32(%0,%%rax), %%xmm2	\n\t"
            "movntdqa 	48(%0,%%rax), %%xmm3	\n\t"
            "movntdqa 	64(%0,%%rax), %%xmm4	\n\t"
            "movntdqa 	80(%0,%%rax), %%xmm5	\n\t"
            "movntdqa 	96(%0,%%rax), %%xmm6	\n\t"
            "movntdqa 	112(%0,%%rax), %%xmm7	\n\t"

            "movntps	%%xmm0, (%1,%%rax)	\n\t"
            "movntps	%%xmm1, 16(%1,%%rax)	\n\t"
            "movntps	%%xmm2, 32(%1,%%rax)	\n\t"
            "movntps	%%xmm3, 48(%1,%%rax)	\n\t"
            "movntps	%%xmm4, 64(%1,%%rax)	\n\t"
            "movntps	%%xmm5, 80(%1,%%rax)	\n\t"
            "movntps	%%xmm6, 96(%1,%%rax)	\n\t"
            "movntps	%%xmm7, 112(%1,%%rax)	\n\t"

            "add	$128, %%rax		\n\t"
            "dec	%2			\n\t"
            "jnz 	1b			\n\t"
            "pop 	%2			\n\t"

            "sfence"
    :
    : "p" (dst), "p" (src), "r" (sse_size)
    : "%rax"
        );

}

void fast_memcpy_ntdqa_ntdq(void *dst, void *src, size_t size) {
    size_t sse_size = (size / 128);

    __asm__ __volatile__ (
            "push 	%2			\n\t"
            "mov        $0, %%rax		\n\t"

            "1:					\n\t"
            "prefetchnta 128(%0,%%rax)		\n\t"
            "prefetchnta 160(%0,%%rax)		\n\t"
            "prefetchnta 192(%0,%%rax)		\n\t"
            "prefetchnta 224(%0,%%rax)		\n\t"


            "movntdqa 	(%0,%%rax), %%xmm0	\n\t"
            "movntdqa 	16(%0,%%rax), %%xmm1	\n\t"
            "movntdqa 	32(%0,%%rax), %%xmm2	\n\t"
            "movntdqa 	48(%0,%%rax), %%xmm3	\n\t"
            "movntdqa 	64(%0,%%rax), %%xmm4	\n\t"
            "movntdqa 	80(%0,%%rax), %%xmm5	\n\t"
            "movntdqa 	96(%0,%%rax), %%xmm6	\n\t"
            "movntdqa 	112(%0,%%rax), %%xmm7	\n\t"

            "movntdq	%%xmm0, (%1,%%rax)	\n\t"
            "movntdq	%%xmm1, 16(%1,%%rax)	\n\t"
            "movntdq	%%xmm2, 32(%1,%%rax)	\n\t"
            "movntdq	%%xmm3, 48(%1,%%rax)	\n\t"
            "movntdq	%%xmm4, 64(%1,%%rax)	\n\t"
            "movntdq	%%xmm5, 80(%1,%%rax)	\n\t"
            "movntdq	%%xmm6, 96(%1,%%rax)	\n\t"
            "movntdq	%%xmm7, 112(%1,%%rax)	\n\t"

            "add	$128, %%rax		\n\t"
            "dec	%2			\n\t"
            "jnz 	1b			\n\t"
            "pop 	%2			\n\t"

            "sfence"
    :
    : "p" (dst), "p" (src), "r" (sse_size)
    : "%rax"
        );
}


void fast_memcpy_aps(void *dst, void *src, size_t size) {
    size_t sse_size = (size / 128);

    __asm__ __volatile__ (
            "push 	%2			\n\t"
            "mov        $0, %%rax		\n\t"

            "1:					\n\t"
            "prefetchnta 128(%0,%%rax)		\n\t"
            "prefetchnta 160(%0,%%rax)		\n\t"
            "prefetchnta 192(%0,%%rax)		\n\t"
            "prefetchnta 224(%0,%%rax)		\n\t"


            "movaps 	(%0,%%rax), %%xmm0	\n\t"
            "movaps 	16(%0,%%rax), %%xmm1	\n\t"
            "movaps 	32(%0,%%rax), %%xmm2	\n\t"
            "movaps 	48(%0,%%rax), %%xmm3	\n\t"
            "movaps 	64(%0,%%rax), %%xmm4	\n\t"
            "movaps 	80(%0,%%rax), %%xmm5	\n\t"
            "movaps 	96(%0,%%rax), %%xmm6	\n\t"
            "movaps 	112(%0,%%rax), %%xmm7	\n\t"

            "movaps	%%xmm0, (%1,%%rax)	\n\t"
            "movaps	%%xmm1, 16(%1,%%rax)	\n\t"
            "movaps	%%xmm2, 32(%1,%%rax)	\n\t"
            "movaps	%%xmm3, 48(%1,%%rax)	\n\t"
            "movaps	%%xmm4, 64(%1,%%rax)	\n\t"
            "movaps	%%xmm5, 80(%1,%%rax)	\n\t"
            "movaps	%%xmm6, 96(%1,%%rax)	\n\t"
            "movaps	%%xmm7, 112(%1,%%rax)	\n\t"

            "add	$128, %%rax		\n\t"
            "dec	%2			\n\t"
            "jnz 	1b			\n\t"
            "pop 	%2			\n\t"

            "sfence"
    :
    : "p" (dst), "p" (src), "r" (sse_size)
    : "%rax"
        );
}

void fast_memcpy_avx(void *dst, void *src, size_t size) {
    size_t sse_size = (size / 512);

    __asm__ __volatile__ (
            "push 	%2			\n\t"
            "mov        $0, %%rax		\n\t"

            "1:					\n\t"
/*            "prefetchnta 128(%0,%%rax)		\n\t"
            "prefetchnta 160(%0,%%rax)		\n\t"
            "prefetchnta 192(%0,%%rax)		\n\t"
            "prefetchnta 224(%0,%%rax)		\n\t"*/


            "vmovdqa 	   (%0,%%rax), %%ymm0	\n\t"
            "vmovdqa 	 32(%0,%%rax), %%ymm1	\n\t"
            "vmovdqa 	 64(%0,%%rax), %%ymm2	\n\t"
            "vmovdqa 	 96(%0,%%rax), %%ymm3	\n\t"
            "vmovdqa 	128(%0,%%rax), %%ymm4	\n\t"
            "vmovdqa 	160(%0,%%rax), %%ymm5	\n\t"
            "vmovdqa 	192(%0,%%rax), %%ymm6	\n\t"
            "vmovdqa 	224(%0,%%rax), %%ymm7	\n\t"

            "vmovdqa 	256(%0,%%rax), %%ymm8	\n\t"
            "vmovdqa 	288(%0,%%rax), %%ymm9	\n\t"
            "vmovdqa 	320(%0,%%rax), %%ymm10	\n\t"
            "vmovdqa 	352(%0,%%rax), %%ymm11	\n\t"
            "vmovdqa 	384(%0,%%rax), %%ymm12	\n\t"
            "vmovdqa 	416(%0,%%rax), %%ymm13	\n\t"
            "vmovdqa 	448(%0,%%rax), %%ymm14	\n\t"
            "vmovdqa 	480(%0,%%rax), %%ymm15	\n\t"

            "vmovntps	%%ymm0,    (%1,%%rax)	\n\t"
            "vmovntps	%%ymm1,  32(%1,%%rax)	\n\t"
            "vmovntps	%%ymm2,  64(%1,%%rax)	\n\t"
            "vmovntps	%%ymm3,  96(%1,%%rax)	\n\t"
            "vmovntps	%%ymm4, 128(%1,%%rax)	\n\t"
            "vmovntps	%%ymm5, 160(%1,%%rax)	\n\t"
            "vmovntps	%%ymm6, 192(%1,%%rax)	\n\t"
            "vmovntps	%%ymm7, 224(%1,%%rax)	\n\t"

            "vmovntps	%%ymm8,  256(%1,%%rax)	\n\t"
            "vmovntps	%%ymm9,  288(%1,%%rax)	\n\t"
            "vmovntps	%%ymm10, 320(%1,%%rax)	\n\t"
            "vmovntps	%%ymm11, 352(%1,%%rax)	\n\t"
            "vmovntps	%%ymm12, 384(%1,%%rax)	\n\t"
            "vmovntps	%%ymm13, 416(%1,%%rax)	\n\t"
            "vmovntps	%%ymm14, 448(%1,%%rax)	\n\t"
            "vmovntps	%%ymm15, 480(%1,%%rax)	\n\t"

            "add	$512, %%rax		\n\t"
            "dec	%2			\n\t"
            "jnz 	1b			\n\t"
            "pop 	%2			\n\t"

            "sfence"
    :
    : "p" (dst), "p" (src), "r" (sse_size)
    : "%rax"
        );
}

void fast_memcpy_avx_multipage(void *dst, void *src, size_t size) {
    size_t page_size = (4096 / 64);
    size_t sse_size = (size / 4096 / 8);

    __asm__ __volatile__ (
            "push 	%2			\n\t"
            "push 	%3			\n\t"
            "mov        $0, %%rax		\n\t"
            "mov        $4096, %%rbx		\n\t"
            "mov        $8192, %%rcx		\n\t"
            "mov        $12288, %%rdx		\n\t"
            "mov        $16384, %%r8		\n\t"
            "mov        $20480, %%r9		\n\t"
            "mov        $24576, %%r10		\n\t"
            "mov        $28672, %%r11		\n\t"

            "1:					\n\t"
            "vmovdqa 	   (%0,%%rax), %%ymm0	\n\t"
            "vmovdqa 	 32(%0,%%rax), %%ymm1	\n\t"
            "vmovdqa 	   (%0,%%rbx), %%ymm2	\n\t"
            "vmovdqa 	 32(%0,%%rbx), %%ymm3	\n\t"

	    "vmovdqa 	   (%0,%%rcx), %%ymm4	\n\t"
            "vmovdqa 	 32(%0,%%rcx), %%ymm5	\n\t"
            "vmovdqa 	   (%0,%%rdx), %%ymm6	\n\t"
            "vmovdqa 	 32(%0,%%rdx), %%ymm7	\n\t"

            "vmovdqa 	   (%0,%%r8), %%ymm8	\n\t"
            "vmovdqa 	 32(%0,%%r8), %%ymm9	\n\t"
            "vmovdqa 	   (%0,%%r9), %%ymm10	\n\t"
            "vmovdqa 	 32(%0,%%r9), %%ymm11	\n\t"

            "vmovdqa 	   (%0,%%r10), %%ymm12	\n\t"
            "vmovdqa 	 32(%0,%%r10), %%ymm13	\n\t"
            "vmovdqa 	   (%0,%%r11), %%ymm14	\n\t"
            "vmovdqa 	 32(%0,%%r11), %%ymm15	\n\t"


            "vmovntps	%%ymm0,    (%1,%%rax)	\n\t"
            "vmovntps	%%ymm1,  32(%1,%%rax)	\n\t"
            "vmovntps	%%ymm2,    (%1,%%rbx)	\n\t"
            "vmovntps	%%ymm3,  32(%1,%%rbx)	\n\t"

            "vmovntps	%%ymm4,    (%1,%%rcx)	\n\t"
            "vmovntps	%%ymm5,  32(%1,%%rcx)	\n\t"
            "vmovntps	%%ymm6,    (%1,%%rdx)	\n\t"
            "vmovntps	%%ymm7,  32(%1,%%rdx)	\n\t"

            "vmovntps	%%ymm8,    (%1,%%r8)	\n\t"
            "vmovntps	%%ymm9,  32(%1,%%r8)	\n\t"
            "vmovntps	%%ymm10,   (%1,%%r9)	\n\t"
            "vmovntps	%%ymm11, 32(%1,%%r9)	\n\t"

            "vmovntps	%%ymm12,    (%1,%%r10)	\n\t"
            "vmovntps	%%ymm13,  32(%1,%%r10)	\n\t"
            "vmovntps	%%ymm14,    (%1,%%r11)	\n\t"
            "vmovntps	%%ymm15,  32(%1,%%r11)	\n\t"

            "add	$64, %%rax		\n\t"
            "add	$64, %%rbx		\n\t"
            "add	$64, %%rcx		\n\t"
            "add	$64, %%rdx		\n\t"
            "add	$64, %%r8		\n\t"
            "add	$64, %%r9		\n\t"
            "add	$64, %%r10		\n\t"
            "add	$64, %%r11		\n\t"
            "dec	%2			\n\t"
            "jnz 	1b			\n\t"
	    "mov	$64, %2			\n\t"
            "add	$4096, %%rax		\n\t"
            "add	$4096, %%rbx		\n\t"
            "add	$4096, %%rcx		\n\t"
            "add	$4096, %%rdx		\n\t"
            "add	$4096, %%r8		\n\t"
            "add	$4096, %%r9		\n\t"
            "add	$4096, %%r10		\n\t"
            "add	$4096, %%r11		\n\t"
            "dec	%3			\n\t"
            "jnz 	1b			\n\t"
            "pop 	%3			\n\t"
            "pop 	%2			\n\t"

            "sfence"
    :
    : "p" (dst), "p" (src), "r" (page_size), "r" (sse_size)
    : "%rax", "%rbx", "%rcx", "%rdx", "%r8", "%r9", "%r10", "%r11"
        );
}
#endif /* GNU */

typedef void (*malloc_function)(void **data, size_t size);
typedef void (*memcpy_function)(void *dst, void *src, size_t size);
typedef void (*free_function)(void* data, size_t size);

struct test_t {
    size_t size;

    malloc_function src_malloc;
    malloc_function dst_malloc;
    memcpy_function memcpy;
    free_function src_free;
    free_function dst_free;
    
    int ready, start, done;
};


void *memcpy_thread(volatile struct test_t *ctx) {
    int i;
    void *src, *dst;

    ctx->src_malloc(&src, ctx->size);
    ctx->dst_malloc(&dst, ctx->size);
    
    ctx->ready = 1;
    while (!ctx->start);
    
    for (i = 0; i < ITERS; i++) {
	ctx->memcpy(dst, src, ctx->size);
    }
    ctx->done = 1;
    
    ctx->dst_free(dst, ctx->size);
    ctx->src_free(src, ctx->size);
    return NULL;
}

void run_memcpy(size_t threads, struct test_t *test, struct timeval *start, struct timeval *end) {
    int err;
    size_t i;
    volatile struct test_t ctx[threads];
    pthread_t th[threads];

    for (i = 0; i < threads; i++) {
	memcpy((void*)&ctx[i], test, sizeof(struct test_t));
	err = pthread_create(&th[i], NULL, (void *(*)(void*))&memcpy_thread, (void*)&ctx[i]);
	if (err) perror("pthread failed");
    }

    for (i = 0; i < threads; i++) {
	while (!ctx[i].ready);
    }

    for (i = 0; i < threads; i++) {
	ctx[i].start = 1;
    }

    gettimeofday(start, NULL);

    for (i = 0; i < threads; i++) {
	while (!ctx[i].done);
    }

    gettimeofday(end, NULL);

    for (i = 0; i < threads; i++) {
	pthread_join(th[i], NULL);
    }
}

int hw_get_cpu_count(void) {
    int err;

    int cpu_count;
    cpu_set_t mask;

    err = sched_getaffinity(getpid(), sizeof(mask), &mask);
    if (err) return 1;

# ifdef CPU_COUNT
    cpu_count = CPU_COUNT(&mask);
# else
    for (cpu_count = 0; cpu_count < CPU_SETSIZE; cpu_count++) {
	if (!CPU_ISSET(cpu_count, &mask)) break;
    }
# endif

    return cpu_count;
}


void ALLOC_std(void **a, size_t size) {
    posix_memalign(a, 4096, size);
}

void FREE_std(void *a, size_t size) {
    free(a);
}

void ALLOC_mmap(void **a, size_t size) {
    *a  =  mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS | MAP_HUGETLB | MAP_LOCKED | MAP_NORESERVE | MAP_POPULATE, -1, 0);
}

void FREE_mmap(void *a, size_t size) {
    munmap(a, size);
}


#ifdef USE_CUDA
void ALLOC_cuda(void **a, size_t size) {
    cudaHostAlloc(a, SIZE, cudaHostAllocPortable);
}

void FREE_cuda(void *a, size_t size) {
    cudaFreeHost(a);
}

void ALLOC_cudawc(void **a, size_t size) {
    cudaHostAlloc(a, SIZE, cudaHostAllocPortable|cudaHostAllocWriteCombined);
}

void FREE_cudawc(void *a, size_t size) {
    cudaFreeHost(a);
}
#endif /* USE_CUDA */

#define PTEST(memtest, mem_in, mem_out, count) \
    memset(&test, 0, sizeof(struct test_t)); \
    test.size = SIZE; \
    test.src_malloc = ALLOC_##mem_in; \
    test.dst_malloc = ALLOC_##mem_out; \
    test.src_free = FREE_##mem_in; \
    test.dst_free = &FREE_##mem_out; \
    test.memcpy = fast_memcpy_##memtest; \
    run_memcpy(count, &test, &tv1, &tv2); \
    printf("Test %s (%u threads, %s => %s): %.2lf GB/s\n", #memtest, count, #mem_in, #mem_out, 1000000. * count * SIZE * ITERS / ((tv2.tv_sec - tv1.tv_sec) * 1000000 + tv2.tv_usec - tv1.tv_usec) / (1024 * 1024 * 1024));

#ifdef THREADS
# if THREADS > 1
#  define PTESTS(test, mem_in, mem_out, count) \
    PTEST(test, mem_in, mem_out, THREADS)
# else /* all */
#  define PTESTS(test, mem_in, mem_out, count) \
    for (i = 2; i < cpus; i *= 2) { \
	PTEST(test, mem_in, mem_out, i) \
    } \
    if ((i != cpus)&&((cpus%2)==0)) { \
	PTEST(test, mem_in, mem_out, cpus/2) \
    } \
    PTEST(test, mem_in, mem_out, cpus)
# endif /* THREADS count */
#else
# define PTESTS(test, mem_in, mem_out, count)
#endif /* THREADS */

void mptest(memcpy_function test, void *dst, void *src, size_t size) {
    int i;

#pragma omp parallel for
    for (i = 0; i < ITERS; i++) {
	test(dst + i * size, src + i * size, size);
    }
}

#ifdef MULTIMODE
# define MPTEST(test, mem_in, mem_out) \
    ALLOC_##mem_in(&a, SIZE * ITERS); \
    ALLOC_##mem_out(&b, SIZE * ITERS); \
    gettimeofday(&tv1, NULL); \
    mptest(fast_memcpy_##test, a, b, SIZE); \
    gettimeofday(&tv2, NULL); \
    FREE_##mem_out(b, SIZE * ITERS); \
    FREE_##mem_in(a, SIZE * ITERS); \
    printf("Test %s (%s => %s): %.2lf GB/s\n", #test, #mem_in, #mem_out, 1000000. * SIZE * ITERS / ((tv2.tv_sec - tv1.tv_sec) * 1000000 + tv2.tv_usec - tv1.tv_usec) / (1024 * 1024 * 1024));\
    PTESTS(test, mem_in, mem_out, cpus)
#else
# define MPTEST(test, mem_in, mem_out)
#endif /* MULTIMODE */

#define TEST(test, mem_in, mem_out) \
    ALLOC_##mem_in(&a, SIZE); \
    ALLOC_##mem_out(&b, SIZE); \
    gettimeofday(&tv1, NULL); \
    for (i = 0; i < ITERS; i++) { \
        fast_memcpy_##test(a, b, SIZE); \
    } \
    gettimeofday(&tv2, NULL); \
    FREE_##mem_out(b, SIZE); \
    FREE_##mem_in(a, SIZE); \
    printf("Test %s (%s => %s): %.2lf GB/s\n", #test, #mem_in, #mem_out, 1000000. * SIZE * ITERS / ((tv2.tv_sec - tv1.tv_sec) * 1000000 + tv2.tv_usec - tv1.tv_usec) / (1024 * 1024 * 1024));\
    MPTEST(test, mem_in, mem_out) \
    PTESTS(test, mem_in, mem_out, cpus)


#ifdef USE_CUDA
# define TESTS(test) \
    TEST(test, std, std) \
    TEST(test, mmap, mmap) \
    TEST(test, cuda, cuda) \
    TEST(test, cudawc, cudawc) \
    TEST(test, std, cuda) \
    TEST(test, cuda, std) \
    TEST(test, std, cudawc) \
    TEST(test, cudawc, std) \
    TEST(test, cuda, cudawc) \
    TEST(test, cudawc, cuda)
#else
# define TESTS(test) \
    TEST(test, std, std) \
    TEST(test, mmap, mmap)
#endif /* USE_CUDA */

int main() {
    int i;
    size_t cpus = hw_get_cpu_count();
    struct test_t test;
    struct timeval tv1, tv2;
    void *a, *b;

#ifdef USE_CUDA
    cudaHostAlloc(&a, SIZE, cudaHostAllocPortable);//4);//cudaHostAllocWriteCombined);
#endif /* USE_CUDA */

/*    
//    TESTS(gcc)
    TESTS(glibc)
//    TESTS(glibc_blocked)
//    TESTS(bcopy)
    TESTS(vik)
*/

#ifndef __INTEL_COMPILER
//    TESTS(intrinsic)
/*    TESTS(repmovs)
    TESTS(dqa_ntps)
    TESTS(ntdqa_ntps)
    TESTS(ntdqa_ntdq)
    TESTS(aps)*/
    TESTS(avx_multipage)
//    TESTS(avx)
#endif /* GNU */

}
